#!/usr/bin/env python
#
# Raccoon
#
# Tk Virtual Screening Interface for AutoDock
#
# v.1.1.0  Stefano Forli
#
#            Copyright 2012, Molecular Graphics Lab
#                The Scripps Research Institute
#                   _  
#                  (,)  T  h e
#                 _/
#                (.)    S  c r i p p s
#                 '\_
#                  (,)  R  e s e a r c h
#                 ./'
#                ( )    I  n s t i t u t e
#                 "
#
#
#     This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>
#
#
#################################################################
#
# Hofstadter's Law: It always takes longer than you expect,
#                   even when you take Hofstadter's Law 
#                   into account.
#
#    The Guide is definitive. Reality is frequently inaccurate.
#        Douglas Adams 
#
#################################################################
#
#
# day_1:   Monday, February 6 2012
# day_2:   Monday, February 13 2012
# day_3:   Tuesday, February 14 2012
# day_4:   Wednesday, February 15 2012
# day_X:   Wednesday, February 22 2012  first working GPF generator
# day_X+1: Friday, February 24 2012 first support for flex residues, 
#                                   first totally successful updateStatus()
#          Tuesday, February 29 2012      first complete cycle, 2 ligs, 2 Rec_rigid, 1 Rec_flex, GPF, DPF(sort of)
#          Wednesday, March 1 2012      first complete running cycle 5K ligs, 2 Rec_rigid, 1 Rec_flex, GPF, DPF
#



import platform # system, platform.uname()
import os, errno # error number for handling error
from sys import argv, exc_info
from HelperFunctionsN3P import *
import shutil
#import Tkinter.Photo



class Raccoon:
    def __init__ (self, mode='autodock', debug = False):

        # AutoDock, Vina modes
        self.mode = mode # 'autodock', 'vina'
        self.debug=debug

        # System identification
        self.system_info = platform.uname()
        if self.debug: print "Raccoon> system info:", self.system_info
        self.platform = self.system_info[0]

        # checking that the right modules are in place.

        # MolKit stuff
        try:
            # from the prepare_ligand code
            from MolKit import Read
            from AutoDockTools.MoleculePreparation import AD4LigandPreparation, AD4ReceptorPreparation, AD4FlexibleReceptorPreparation
            # from the prepare_flexres code
            from MolKit.protein import ProteinSet, ResidueSet
            from MolKit.molecule import BondSet
            from MolKit.stringSelector import CompoundStringSelector
            from AutoDockTools.MoleculePreparation import AD4FlexibleReceptorPreparation
        except:
            error_title = "MolKit error!"
            error_msg = "Impossible to find the MolKit module."
            error_msg += "\n\nMGLTools is required to run Raccoon but is either misconfigured or not installed"
            if not self.platform == "Windows":
                suggestion = "\n\nPlease install it or be sure to use the right Python interpreter."
            else:
                suggestion = "\n\nPlease install it, or try to run Raccoon with:\n\n $MGLROOT/bin/pythonsh raccoon.py"
            print error_title, error_msg, suggestion
            exit(1)    

        self.hcovbond = 1.1
        self.acceptedFiles = ['pdbqt', 'pdb', 'mol2']

        self.initLigData()
        self.initRecData()
        self.initLigFilters(mode=self.mode) # provide self.filterSettings
        self.initCommonSettings() 
        self.initAutoDockGpfSettings()
        self.initAutoDockDpfSettings()
        self.initVinaSettings()
        self.initSystemOptions()

    def initCommonSettings(self):
        """ initialize common settings for AutoDock and Vina """
        # search space center
        self.box_center = [ None, None, None ]
        # search space dimensions
        self.box_size = [ None, None, None ]
        # output directory
        self.output_dir = { 'path' : None,     # full path
                            'freespace': None, # available disk space (used for predictions?)
                            'used' : False,    # flag to tag the dir once used in a (partial?) generation
                             }

        self.status = { 'lig': False,          # variables for keeping
                        'rec': False,          # track of Raccoon status 
                        'grid_box': False,     # and when the generation
                        'docking': False,      # is allowed
                        'directory': False}

    def initAutoDockGpfSettings(self):
        """ initialize settings for AutoGrid """

        self.autogrid_parameters = { 'gridcenter': [ None, None, None ],
                                    'npts' : [ None, None, None ],
                                    'parameter_file' : None,
                                    'spacing': 0.375,
                                    'receptor_types' : [],
                                    'ligand_types': [],
                                    'receptor': '',
                                    'gridfld' : '',
                                    'smooth' : 0.5,
                                    'map' : [],
                                    'dielectric': -0.1456,
                                    'KEYWORD_ORDER' : [ 'npts', 'gridfld', 'spacing', 'receptor_types', 
                                                        'ligand_types', 'receptor','gridcenter',
                                                        'smooth', 'map', 'dielectric'] }
        ## raccoon options
        #
        # map mode
        self.autogrid_mode = 'always' # 'now' or 'never' (C) 1960 Elvis
        # map copy/link
        self.autogrid_map_file = 'copy' # 'link'
        # binary to use for calculations
        self.autogrid_bin = None # will be initialized with a 'which'
        # map files per atom types
        self.autogrid_maps = {} # { 'A': 'protein.A.map', 'C' : 'protein.C.map', 


    def initAutoDockDpfSettings(self):
        """ initialize settings for AutoDock """

        self.autodock_parameters = {


        'search_mode'   : 'lga', # lga, ga, sa, ls
        'runs'          :  100,  # common to all search methods
        'do_clustering' : True,  # perform clustering
        'complexity_map': None,  # map used to calculate ligand accessibility (usually *.C.map)
        'dpf_mode'      : 'template', # dpf can be generated from'template' or 'auto' (complexity map)


        # generic common settings
        'generic' : {
            'autodock_parameter_version' : '4.2',
            'parameter_file' : None,
            'outlev' : 0,
            'intelec': True,
            'seed' : [ 'pid time'],
            'unbound_model' : 'bound',
            'KEYWORD_ORDER'  : ['outlev','intelec', 'seed', 'unbound_model' ],
            'OPTIONAL'       : [],
            },

        # ligand specific settings
        'ligand' : { 'ligand_types' : [],
                   'map' : [],
                   'move' : '',
                   'flexres' : '',
                   'about' : [0.0, 0.0, 0.0], 
                   'KEYWORD_ORDER' : [ 'ligand_types', 'map', 
                        'fld', 'elecmap', 'desolvmap', 'move', 'flexres', 'about'],
                   'OPTIONAL' : []
            },

        # genetic algorithm search settings
        'ga'     : {
            'ga_pop_size' : 150,
            'ga_num_evals': 2500000,
            'ga_num_generations':   27000,
            'ga_elitism'    : 1,
            'ga_mutation_rate': 0.02,
            'ga_crossover_rate': 0.8,
            'ga_window_size' : 10,
            'ga_cauchy_alpha': 0.0,
            'ga_cauchy_beta': 1.0,
            'set_ga'    : '',
            'KEYWORD_ORDER':['ga_pop_size','ga_num_evals', 
                             'ga_num_generations', 'ga_elitism', 
                             'ga_mutation_rate', 'ga_crossover_rate',
                             'ga_window_size', 'ga_cauchy_alpha',
                             'ga_cauchy_beta', 'set_ga'],
            'OPTIONAL'     :[],
            },

        # simulated annealing search settings
        'sa': { 'tstep': 2.0,
                'qstep': 50.0,
                'dstep': 50.0,
                'rt0':   1000.0,
                'scheduler': 'linear_schedule', # 'linear_schedule', None
                'rtrf': 0.95,
                'trnrf': 1.0,
                'quarf': 1.0,
                'dihrf': 1.0,
                'cycles': 50,
                'accs': 100,
                'rejs': 100,
                'select': 'm',
                'KEYWORD_ORDER': [ 'tstep', 'qstep', 'dstep', 'rt0', 'scheduler',
                                  'rtrf', 'trnrf', 'quarf', 'dihrf', 'cycles', 'accs',
                                  'rejs', 'select']
               },

        # local search settings
        'ls' : { 
            'ga_pop_size' : 150,
            'sw_max_its': 300,
            'sw_max_succ' : 4,
            'sw_max_fail' : 4,
            'sw_rho'    : 1.0,
            'sw_lb_rho' : 0.01,
            'ls_search_freq': 0.06,
            'set_psw1' : '',
            'KEYWORD_ORDER' : [ 'ga_pop_size', 'sw_max_its', 'sw_max_succ', 'sw_max_fail',
                               'sw_rho', 'sw_lb_rho', 'ls_search_freq', 
                               'set_psw1'],
            },

        # clustering analysis settings
        'clustering': {
            'rmstol' : 2.0,
            'analysis' : '',
            'KEYWORD_ORDER': ['rmstol', 'analysis']
            },
                # activate GA           : 'ga_run'
                # activate LS             : 'set_psw1', 'set_sw1'
                # activate Clustering   : 'analysis'
                # ACTIVATION ORDER: generic, ls, ga, clustering

            } # end autodocking settings
    
        # XXX unbound energies... calculate?
        #  LS = (ga_pop_size + ls_block + do_local_only XX + analysis )
        # LGA = (ga_block + ls_block + ga_run XX + analysis)
        # Refer to "self.autodock_runs" value

    def initVinaSettings(self):
        # vina data
        self.vina_settings = { 
               'center_x': 0.0,
               'center_y': 0.0,
               'center_z': 0.0,
               'size_x'  : 0.0,
               'size_y'  : 0.0,
               'size_z'  : 0.0,
               'out'     : '',
               'log'     : '',
               'cpu'     : 1,
               'exhaustiveness': 8,
               'num_modes' : 9,
               'energy_range': 3.,
               'KEYWORD_ORDER': [ 'center_x', 'center_y', 'center_z',
                                  'size_x', 'size_y', 'size_z',
                                  'exhaustiveness', 'cpu', 'out', 'log',
                                  'num_modes', 'energy_range',
                                ],
               'OPTIONAL' : {'out': False, 'log' : False, 'num_modes' : False, 'energy_range': False }, 
        }



    def initSystemOptions(self):
        """ initialize general options 'self.generalOpts', used 
            during the generation of the jobs
        """

        self.generalOpts = { 
           'generic': {  'computing'            : 'workstation', # workstation, linux_cluster
                         'platform'             : 'current', # current (sys.platform), 'posix', 'windows', 'cygwin'
                         'split'                : 1000,  # None, 10, 100, 1000, 10K 100K
                         'script'               : 'master', # master, single, off
                         'package'              : 'off', # tar, targz, tarbz2, zip, off
                         'delete_maps'          : True, # true/false
                         'out_file_per_ligand'  : 1,  # 1  
                         'cpu_per_job'          : 1,  # 1, 2, 3
                      },

           'scheduler': 'pbs', # currently, only 'pbs'; FUTURE: sungrid, opal ...?

           'scheduler_pbs': { 
                         'cput'        : '24:00:00',
                         'walltime'    : '24:00:00',
                         'nodes'       : 1,
                         'ppn'         : 1,
                         'mem'         : '512mb',
                         'scratch_dir' : '',
                         'qsub_bin'    : 'qsub',
                         'autodock_bin': 'autodock4',
                         'autogrid_bin': 'autogrid4',
                         'vina_bin'    : 'vina',
                         'options'     : '', #OPT="-q MyPriorQueue"
                    },
            }



    def setMode(self, mode ='autodock'):
        """handle all the changes to be made to Raccoon
            to handle the different engines (autodock,vina)
        """
        # if current settings are previous mode default
        # this setting must be updated
        self.mode = mode
        if self.filterSettings == self.FILTER_DEF:
            current_default = True
        else:
            current_default = False
        self.initLigFilters(mode=mode)

        if current_default: 
            self.filterSettings = self.FILTER_DEF.copy()
        # XXX incomplete... XXX


    def setBoxCenter(self, coord):
        """ accepts a vector with 3 coordinates (A)"""
        try:
            self.box_center = [ float(coord[0]), float(coord[1]), float(coord[2]) ]
            # update the vina config template
            self.vina_settings['center_x'] = self.box_center[0]
            self.vina_settings['center_y'] = self.box_center[1]
            self.vina_settings['center_z'] = self.box_center[2]
            # update the autodock config template
            self.autogrid_parameters['gridcenter'] = self.box_center[:]

            return True
        except:
            if self.debug:
                print "setBoxCenter> ERROR in setting coords(%s) : %s" % (coord, exc_info()[1]  )
            return False 
            

    def setBoxSize(self, size):
        """ accepts a vector with three dimensions (A)
            numbers are adapted later on to be points 
            (AD) or actual angstroms (VINA)
        """
        try:
            self.box_size = [ float(size[0]), float(size[1]), float(size[2]) ]
            # update the vina config template
            self.vina_settings['size_x'] = self.box_size[0]
            self.vina_settings['size_y'] = self.box_size[1]
            self.vina_settings['size_z'] = self.box_size[2]
            # update the autogrid config template
            self.autogrid_parameters['npts'] = [ int(x/self.autogrid_parameters['spacing'])+1 for x in self.box_size ]
            return True
        except:
            if self.debug:
                print "setBoxSize> ERROR in setting size (%s) : %s" % (size, exc_info()[1]  )
            return False

    def setAutoGridResolution(self, spacing):
        try:
            spacing = float(spacing)
        except:
            if self.debug: print "setAutoGridResolution> ERROR setting grid resolution [%s]: %s" % (spacing, exc_info()[1])
            return False
        self.autogrid_parameters['spacing'] = spacing
        size = [ x * spacing for x in self.autogrid_parameters['npts'] ]
        self.setBoxSize(size)
        return True


    def setOutputDir(self, dir_path, autoCreate=True, checkEmpty=True):
        try:
            # check if dir exists
            if not os.path.exists(dir_path):
                if not autoCreate:
                    if self.debug: print "setOutputDir> ERROR directory (%s) doesn't exist (autoCreate:False)" % (dir_path)
                    return (False, 'inexistent')
                try:
                    os.makedirs(dir_path)
                except:
                    # XXX add here all teh tests.. /dev/null dev/full
                    if self.debug: print "setOutputDir> ERROR impossible to create dir (%s): %s" % (dir_path, exc_info()[1])  
                    return (False, ('error creating dir %s: %s' % (dir_path, exc_info()[1]) ) )

            # check if dir is empty
            if checkEmpty and not os.listdir(dir_path) == []: 
                if self.debug: print "setOutputDir> ERROR directory not empty (%s) (checkEmpty:True)" % dir_path
                return (False, 'not empty')

            # successful
            self.output_dir['path'] = dir_path
            self.output_dir['freespace'] = checkDiskSpace(dir_path)
            checkDiskSpace(dir_path, human=True)
            self.output_dir['used'] = False
            self.updateStatus()
            return (True, 'no errors')
        except:
            if self.debug: print "setOutputDir> ERROR impossible to access directory (%s) : %s" % (dir_path, exc_info()[1])
            return (False, ('error accessing %s: %s' % (dir_path, exc_info()[1]) ) )

    def initLigData(self):
        """ initialize ligand dictionaries of the Raccoon session"""
        try: del self.LigBook
        except: pass
        self.LigBook = {}

        try: del self.atypelist
        except: pass
        try: del self.unknownAtypes
        except: pass
        # AutoDock atom type list and their count (updated with ligands imported)
        ## Warning: atomic weights are inaccurate and there is no account for merged non-polar H's
        self.defaultAtWeight = 12 # default atomic weight for unknown atype
            

        self.atypelist = {   # count  MW
                    'H'      : [ 0,  1    ],
                    'HD'     : [ 0,  1    ],
                    'HS'     : [ 0,  1    ],
                    'C'      : [ 0,  12   ],
                    'A'      : [ 0,  12   ],
                    'N'      : [ 0,  14   ],
                    'NA'     : [ 0,  14   ],
                    'NS'     : [ 0,  14   ],
                    'OA'     : [ 0,  16   ],
                    'OS'     : [ 0,  16   ],
                    'F'      : [ 0,  19   ],
                    'Mg'     : [ 0,  24   ],
                    'MG'     : [ 0,  24   ],
                    'P'      : [ 0,  31   ],
                    'SA'     : [ 0,  32   ],
                    'S'      : [ 0,  32   ],
                    'Cl'     : [ 0,  35.4 ],
                    'CL'     : [ 0,  35.4 ],
                    'Ca'     : [ 0,  40   ],
                    'CA'     : [ 0,  40   ],
                    'Mn'     : [ 0,  55   ],
                    'MN'     : [ 0,  55   ],
                    'Fe'     : [ 0,  56   ],
                    'FE'     : [ 0,  56   ],
                    'Zn'     : [ 0,  65.4 ],
                    'ZN'     : [ 0,  65.4 ],
                    'Br'     : [ 0,  80   ],
                    'BR'     : [ 0,  80   ],
                    'I'      : [ 0, 126   ],
                                                # XXX remember to put W and G atoms!
                    'e'      : [ 1,   0   ], # always 1 by default XXX check this in the new raccoon
                    'd'      : [ 1,   0   ]  # always 1 by default
                    }
        self.unknownAtypes = {} # 'type' : [ filenames ]


    def initRecData(self):
        """ initialize receptor dictionaries of the Raccoon session"""
        try: del self.RecBook
        except: pass
        self.RecBook = {}

        try: del self.atypelistRec
        except: pass
        self.atypelistRec = {   # count  MW
                    'H'      : [ 0,  1    ],
                    'HD'     : [ 0,  1    ],
                    'HS'     : [ 0,  1    ],
                    'C'      : [ 0,  12   ],
                    'A'      : [ 0,  12   ],
                    'N'      : [ 0,  14   ],
                    'NA'     : [ 0,  14   ],
                    'NS'     : [ 0,  14   ],
                    'OA'     : [ 0,  16   ],
                    'OS'     : [ 0,  16   ],
                    'F'      : [ 0,  19   ],
                    'Mg'     : [ 0,  24   ],
                    'MG'     : [ 0,  24   ],
                    'P'      : [ 0,  31   ],
                    'SA'     : [ 0,  32   ],
                    'S'      : [ 0,  32   ],
                    'Cl'     : [ 0,  35.4 ],
                    'CL'     : [ 0,  35.4 ],
                    'Ca'     : [ 0,  40   ],
                    'CA'     : [ 0,  40   ],
                    'Mn'     : [ 0,  55   ],
                    'MN'     : [ 0,  55   ],
                    'Fe'     : [ 0,  56   ],
                    'FE'     : [ 0,  56   ],
                    'Zn'     : [ 0,  65.4 ],
                    'ZN'     : [ 0,  65.4 ],
                    'Br'     : [ 0,  80   ],
                    'BR'     : [ 0,  80   ],
                    'I'      : [ 0, 126   ],
                                                # XXX remember to put W and G atoms!
                    'e'      : [ 1,   0   ], # always 1 by default XXX check this in the new raccoon
                    'd'      : [ 1,   0   ]  # always 1 by default
                    }

        try: del self.resRotatableBondTable
        except: pass

        # XXX potentially useless... (see self.addReceptor
        self.resRotatableBondTable = {
                'GLY'    : [ 0, [""] ],  # XXX this should raise an error message?
                'ALA'    : [ 0, [""] ],
                'PRO'    : [ 0, [""] ],
                'VAL'    : [ 1, ["C"] ],
                'LEU'    : [ 2, ["C"] ],
                'SER'    : [ 2, ["C", "OA", "HD"] ],
                'THR'    : [ 2, ["C", "OA", "HD"] ],
                'CYS'    : [ 2, ["C", "SA", "HD"] ],
                'ASN'    : [ 2, ["", "", ""] ],
                'PHE'    : [ 2, ["A", "C"] ],
                'TRP'    : [ 2, ["C", "A", "N", "HD"] ],
                'HIE'    : [ 2, ["C", "A", "NA", "N", "HD"] ],
                'HIS'    : [ 2, ["C", "A", "NA", "N", "HD"] ],
                'ASP'    : [ 2, ["C", "OA"] ],
                'ILE'    : [ 2, ["C"] ],
                'GLN'    : [ 3, ["C", "OA","N","HD"] ],
                'TYR'    : [ 3, ["C", "A", "OA", "HD"] ],
                'GLU'    : [ 3, ["C", "OA"] ],
                'MET'    : [ 3, ["C", "S"] ],
                'ARG'    : [ 4, ["C", "N", "HD"] ],
                'LYS'    : [ 5, ["C", "N", "HD"] ] }


    def initLigFilters(self, mode='autodock'):
        """
            create the pre-set filter settings
            (mode-dependent)
        """
        self.autodockMaxRot = 32
        self.vinaMaxRot = 50

        if self.mode == 'autodock':
            self.maxTors = self.autodockMaxRot
        elif self.mode == 'vina':
            self.maxTors = self.vinaMaxRot

        # TODO check if they exist first?


        self.FILTER_DEF = { 'hbd' : [0,99],  'hba' : [0,99],
                            'mw'  : [0,9999],'nat' : [0,999],
                            'tors': [0,self.maxTors], 'accept_unk_atypes': False,
                            'title': 'Default'}
        self.FILTER_DEF_LIPINSKY = { 'hbd' : [0,5],  'hba' : [0,10],
                            'mw'  : [0,500],'nat' : [0,999],
                            'tors': [0,self.maxTors], 'accept_unk_atypes': False,
                            'title': 'Lipinsky-like'}
        self.FILTER_DEF_DRUGLIKE = { 'hbd' : [0,5],  'hba' : [0,10],
                            'mw'  : [160,480],'nat' : [20,70],
                            'tors': [0,self.maxTors], 'accept_unk_atypes': False,
                            'title': 'Drug-like'}
        self.FILTER_DEF_DRUGLIKE_FRAG = { 'hbd' : [0,3],  'hba' : [0,6],
                            'mw'  : [160,300],'nat' : [6,45],
                            'tors': [0,self.maxTors], 'accept_unk_atypes': False,
                            'title': 'Drug-like (fragments)'}

        self.setFilterSet(filterSet='default')


    def setFilterSet(self, filterSet=None, customFilter=None):
        """
            change the current filter set.
            It is possible to specify any of the predefined or
            provide a custom one.

            predefined: default, lipinsky, druglike, druglike_frag

            accepted filter format: { 'hbd': [min,MAX] , 'hba':[min,MAX] ,
                                      'mw' : [min,MAX] , 'nat':[min,MAX] ,
                                      'tors': [min,MAX], 'accept_unk_atype : False,
                                      'title': '' }
        """
        if customFilter:
            setting = customFilter
        else:
            if filterSet == 'default' or filterSet==None:
                setting = self.FILTER_DEF
            elif filterSet == 'lipinsky':
                setting = self.FILTER_DEF_LIPINSKY
            elif filterSet =='druglike':
                setting = self.FILTER_DEF_DRUGLIKE
            elif filterSet =='druglike_frag':
                setting = self.FILTER_DEF_DRUGLIKE_FRAG

        self.filterSettings = setting.copy() # initialize the current filter settings with the default Raccoon

        

    #def receptors(self):
    def receptors(self):
        """ return accepted receptors"""
        out = []
        for r in self.RecBook.keys():
            if self.RecBook[r]['accepted']:
                out.append(r)
        return out

    def currentLigAtypes(self):
        """ 
            return atom types employed in the 
            current session (accepted ligands
        """
        atypes = []
        for a in self.atypelist.keys():
            if self.atypelist[a][0] > 0:
                atypes.append(a)
        return sort(atypes)

    def removeReceptors(self, rec_list=None):
        """ remove receptors from the session"""
        # XXX TODO
        if rec_list == None:
            if self.debug: print "removeReceptors> wyping out everything..."
            self.RecBook.clear()
        else:
            for r in rec_list:
                del self.RecBook[r]

        self.updateStatus()
 
 

    def updateStatus(self, report=False):
        """
        check the status of the session for the generation step:
        - at least one ligand is present and filter-accepted;
        - at least one receptor is present and accepted
        - grid box params are defined (size, position)
        - docking params are defined
        - output dir is defined
        """
        # ligands
        self.status['lig'] = ( len(self.ligands(atLeastOne=True)) > 0 )
        # receptor
        self.status['rec'] = ( len(self.RecBook.keys()) > 0 )
        # grid_box
        center = True
        size = True
        for c in self.box_center:
            if c == None:
                center = False
                #if self.debug: print "MISSING CENTER", self.box_center
                break
        if center:
            for s in self.box_size:
                if s == None:
                    size = False
                    #if self.debug: print "MISSING SIZE", self.box_size
                    break
        self.status['grid_box'] = (center and size)
        # docking
        self.status['docking'] = True
        # directory
        valid_path = False
        if not self.output_dir['path'] == None:
            if not self.output_dir['used']:
                if self.output_dir['freespace'] > 0: # XXX useless placeholder /to be improved
                    valid_path = True    
        self.status['directory'] = valid_path
        # FINAL STATUS
        final_status = (self.status['lig'] and 
                self.status['rec'] and 
                self.status['grid_box'] and 
                self.status['docking'] and 
                self.status['directory'] )

        if report or self.debug:
            print "\nupdateStatus>"
            for i in sorted(self.status.keys()):
                print "\t[   %s\t]\t %s" %(self.status[i], i)
            print '\t---------------------\n\t  FINAL  :\t %s' % final_status 
        return final_status

    def generateLigPdbqt(self, filelist, GUIcallback=None):
        """callback is a function that is called at every step...
        """



    def checkPdbqtList(self, filelist):
        """ 
            check if the file is a valid AutoDock(Vina) file (mode):
                - 'lig' : formatted ligand file
                - 'rec' : target structure file
                - 'flex': formatted flexible residue(s)
            return dictionary { lig, rec, flex, error}
        """
        lig = []
        rec = []
        flex = []
        error = []
        for f in filelist:
            found = None
            try:
                data = getLines(f)
                if len(data) < 1:
                    error.append([f, '(empty)'])
                    continue
                for l in data:
                    if l.startswith('BEGIN_RES'):
                        found = 'flex'
                        flex.append(f)
                        break
                    elif l.startswith('ROOT'):
                        found = 'lig'
                        lig.append(f)
                        break
                    elif l.startswith('ATOM') or l.startswith('HETATM'):
                        found = 'rec'
                        rec.append(f)
                        # if no keywords have been found so far
                        # a PDBQT can be only a rec file
                        break
            except: 
                found = 'file_error [%s]: %s' % (f, exc_info()[1] )
                error.append([f, found])
        return {'lig' : lig, 'rec': rec, 'flex':flex, 'error': error}        


    def addReceptorList(self, receptor_files=None, receptor_flex_pairs_list=None): 
        """ accept a list of PDBQT to be included in receptor's list
            - scan the list
            - reject ligs
            - identify rec, rigid_rec, flex
            - attempt to combine...?

            return a dictionary:
                'accepted' : accepted
                'duplicate': already present
                'rejected' : problematic files

        """
        #print "\n\n\n\n*************************************************"
        #for i in receptor_files:
        #    print "FIOLE>", i



        accepted = []
        rejected = []
        # clean up the list of PDBQT (lig, rec, flex)
        checked_files = self.checkPdbqtList(receptor_files)
        for l in checked_files['lig']:
            rejected.append(['is_ligand', l])
        for fx in checked_files['flex']:
            rejected.append(['is_flexres', fx])
        for e in checked_files['error']:
            rejected.append(['file error', e])

        # find pairs of rec+flex, if any
        rec_flex_pair = self.findRecFlexPairs(mode='name', **checked_files)
        for p in rec_flex_pair:
            r = p[0]
            f = p[1]
            response = self.addReceptor(receptor = r, flex=f)
            if response['accepted']:
                #print "RESPONSE", response
                #item = [response['name'], p]
                #print "ITEM>", item
                accepted.append([response['name'], p])
            else:
                rejected.append([response['reason'], r, f])
        self.updateStatus()
        #print "addReceptorList> ACCEPTED", accepted
        return {'accepted': accepted, 'rejected': rejected}
    
    def findRecFlexPairs(self, mode='name', **files):
        """ from a dictionary of files (generated with checkPdbqtList)
            try to identify if there are rec+flex pairs

            - mode 'name'   ('_rigid', '_flex')
            - mode 'coord' ( 'CA' connected with two atoms)

            WARNING: if '*_rigid.pdbqt' and '*_flex.pdbqt'
                     are generated there is no guarantee 
                     of actual correspondence between the two!

                    Raccoon generation should avoid this problems:
                        
                        1jff.pdbqt -> (B:THR276) -> 1jff_B-THR276_rigid.pdbqt
                                                    1jff_B-THR276_flex.pdbqt

                        1jff.pdbqt -> (B:THR276, -> 1jff_B-THR276_B-HIS229_rigid.pdbqt
                                       B:HIS229)    1jff_B-THR276_B-HIS229_flex.pdbqt
        """
        pairs = []
        if mode == 'name':
            for r in files['rec']:
                rname = os.path.basename(r)
                rname = rname[:-12]
                flex_match = ''
                for f in files['flex']:
                    fname = os.path.basename(f)
                    fname = fname[:-11]
                    if fname == rname:
                        flex_match = f
                        break
                pairs.append( [r, flex_match] )
        elif mode =='coord':
            print "findRecFlexPairs> coord pairing Rec+Flex [ not implemented yet ]"
        return pairs

    def addReceptor(self, receptor, flex=None):
        """ add the receptor to the session"""

        rec_data = self.getRecData(receptor, flex)
        #
        # XXX filter here receptor against allowed atypes/special parm file
        #
        rec_data['accepted'] = True
        #
        name = rec_data['name']
        is_wanted = False
        if not name in self.RecBook.keys():
            is_wanted = True
            reason = 'new'
        else:
            if self.RecBook[name]['filename'] == rec_data['filename']:
                # real duplicate
                reason = 'duplicate'
                is_wanted = False
            else:
                is_wanted = True
                # omonimy
                reason = "omonimy"
                c = 1
                suggest = "%s_v%d" % (name, c)

                while suggest in self.RecBook.keys():
                    c += 1
                    suggest = "%s_v%d" % (name, c)
                    if c == 99:
                        if self.debug: print "addReceptor> too many attempts to find name", suggest
                        is_wanted = False
                        reason = 'no available name up to %s' % suggest
                        break
                if is_wanted: 
                    if self.debug: print ": found suggested name:",suggest
                    rec_data['name'] = suggest
                    name = suggest
        if is_wanted:
            if self.debug: print "REGISTERING", rec_data['name']
            self.registerReceptor(action='add', **rec_data)
        return { 'accepted' : is_wanted, 'reason': reason, 'name': name}

    def getRecData(self, receptor, flex=None):
        """ get receptor data and incorporate flex residues if present"""
        rec_data = self.extractRecProperties(receptor)
        rec_data['is_flexible'] = False
        if flex:
            flex_res_data = self.extractRecProperties(flex)
            rec_data['flex_res_file'] = flex
            rec_data['chains'] = sorted(set(rec_data['chains']+flex_res_data['chains']))
            rec_data['residues'] = sorted(set(rec_data['residues']+flex_res_data['residues']))
            #rec_data['residues'] += flex_res_data['residues']
            rec_data['flex_res'] = sorted(flex_res_data['residues'])
            rec_data['flex_atypes'] = flex_res_data['atypes']
            rec_data['flex_atypes_unknown'] = flex_res_data['atypes_unknown']
            rec_data['flex_tors'] = getPdbqtTors(flex)
            rec_data['is_flexible'] = True
            # C:THR276 -> C-THR276
            flex_names = [ i.replace(":","-") for i in sorted(flex_res_data['residues']) ]
            rec_data['name'] += "_"+"_".join(flex_names) 
        return rec_data

    def generateFlexRes(self):
        # - generate the rigid+flex pair
        # - remove the original FULL_RIGID
        # - import rigid+flex_pair
        pass

    def extractRecProperties(self,filename):
        # -receptor types
        # -residues
        # -residues in the grid?
        name = filetoname(filename)
        data = getLines(filename)
        atoms = getAtoms(data)
        rec_structure = getReceptorResidues(data=atoms)
        rec_types = []
        rec_types_unk = []
        chains = rec_structure.keys()
        res_list = []
        for c in chains:
            res = rec_structure[c]
            for r in res:
                res_list.append( ("%s:%s" % (c.strip(),r)) )
        for a in atoms:          
            at = getAtype(a)
            if at in self.atypelistRec:
                if not at in rec_types:
                    rec_types.append(at)
            else:
                rec_types_unk.append(at)
        return {'name': name,
                'atypes':rec_types,
                'atypes_unknown': rec_types_unk,
                'structure': rec_structure, 
                'chains' : chains,
                'residues': res_list,
                'filename': filename}

    def registerReceptor(self,action='add', **rec_data):
        """ register the receptor in the Great Book of Receptors

        """
        name = rec_data['name']
        if action =='add':
            self.RecBook[name] = rec_data
        elif action =='del':
            del selfRecBook[name]
        self.updateAtomPoolRecUnk(action=action, **rec_data)

    def updateAtomPoolRecUnk(self, action='add', **rec_data):
        """ update receptor unknown atom types """
        #print "updateAtomPoolRecUnk> xxx not working xxx"
        pass
    # XXX probably useless...






    def ligToBox(self, ligand, tolerance=5):
        """ accept a ligand or a AL result (file or lines list) and returns
            the bounding box and the center plus a tolerance (angstroms)
        """
        try:
            #tolerance = 0
            atoms = getAtoms(getLines(ligand))
            lig_bbox = boundingBox(atoms, tolerance)
            self.ligBox = lig_bbox
            return True
        except:
            if self.debug: print "ligToBox> ERROR reading ligand [%s] : [%s]" % (ligand, exc_info()[1])
            return False
            

    def importGpfTemplate(self, gpf):
        """read a GPF file template to set values in the 
           grid box of current session
        """
        try:
            gpf_data = getGPFdata(gpf)
            self.autogrid_parameters['gridcenter'] = gpf_data['center']
            self.autogrid_parameters['npts']       = gpf_data['pts']
            self.autogrid_parameters['spacing']    = gpf_data['res']
            self.autogrid_parameters['smooth']     = gpf_data['smooth']
            self.autogrid_parameters['dielectric'] = gpf_data['dielectric']
            self.gridBoxFromGpf(gpf_data)
            return True
        except:
            print "importGpfTemplate> ERROR reading data file %s: %s" % (gpf, exc_info()[1])
            return False


    def importDpfTemplate(self, dpf):
        """read a DPF file template to set values in the 
           parmeter search
        """
        # XXX INCOMPLETE NOT WORKING!!
        try:
            dpf_data = getGPFdata(gpf)
            self.autogrid_parameters['gridcenter'] = gpf_data['center']
            self.autogrid_parameters['npts']       = gpf_data['pts']
            self.autogrid_parameters['spacing']    = gpf_data['res']
            self.autogrid_parameters['smooth']     = gpf_data['smooth']
            self.autogrid_parameters['dielectric'] = gpf_data['dielectric']
            self.gridBoxFromGpf(gpf_data)
            return True
        except:
            if self.debug: print "importDpfTemplate> ERROR reading data file %s: %s" % (gpf, exc_info()[1])
            retu



    def gridBoxFromGpf(self, gpf_data):
        """set the grid box from an AutoDock GPF """
        size = [ gpf_data['g_max'][0] - gpf_data['g_min'][0],
                 gpf_data['g_max'][1] - gpf_data['g_min'][1],
                 gpf_data['g_max'][2] - gpf_data['g_min'][2] ]

        self.setBoxSize( size )
        self.setBoxCenter( gpf_data['center'])
        return True

    def gridBoxFromConf(self, conf):
        """set the grid box from a Vina config file
           vina conf files don't contain any mandatory parameter,
           therefore a check is provided for missing values 

        """
        # XXX TODO this should change also other keyword settings!
        try:
        #if True:
            conf_data = getVinaConfData(conf)
            center = [ conf_data['center_x'],  conf_data['center_y'],  conf_data['center_z'] ]
            size = [ conf_data['size_x'],  conf_data['size_y'],  conf_data['size_z'] ]
            #print "XXX", center, size
            c = 0
            missing = False
            for v in center:
                if v == None:
                    missing = True
                    break
            if not missing:
                #print self.setBoxCenter, type(self.setBoxCenter)
                self.setBoxCenter( center )
                c +=1
            missing = False
            for v in size:
                if v == None:
                    size = True
                    break
            if not missing:
                self.setBoxSize( size )
                c +=1
            if c==1:
                if self.debug: print "gridBoxFromConf> WARNING incomplete data in file %s " % (conf)
            elif c==0:
                raise "No values found"
            return conf_data

        # else:
        except:
            print "gridBoxFromConf> ERROR reading data file %s : %s" % (conf, exc_info()[1])
            return False

    def gridBox(self):
        return {'center' : self.box_center,
                'size'   : self.box_size,
               }

    def gridBoxFromLigand(self, ligand, tolerance=5):
        """
        used to set grid center and size by using a ligand pdb/qt or 
        an AutoLigand output
        a tolerance can be added (Angstrom)
        """
        try:
            #ligand_lines = getLines(ligand)
    
            self.ligToBox(ligand, tolerance=tolerance)

            self.ligBox['file'] = ligand

            size = [ self.ligBox['x'][1] -  self.ligBox['x'][0],
                     self.ligBox['y'][1] -  self.ligBox['y'][0],
                     self.ligBox['z'][1] -  self.ligBox['z'][0] ]
    
            self.setBoxSize( size )
            self.setBoxCenter( self.ligBox['center'] ) 
        except:
            if self.debug: print "gridBoxFromLigand> ERROR reading ligand [%s] : %s" % (ligand, exc_info()[1])
            return False


    def getVinaSettingsFromConf(self,config):
        """ extract search and parms settings from a config file

        """
        accepted = [ 'exhaustiveness', 'num_modes', 'energy_range' ]

        for k in accepted:    
            if k in conf_data.keys():
                self.vina_settings[k] = conf_data[k]

    def AutoDockMakeDpf(self, ligand, receptor, settings = None):
        ''' build the dpf for ligand+receptor using the
            common settings in self.autodock_parameters{}

            the dpf is build in a modular fashion:
                    generic
                    ligand
                    search(ga,sa)
                    local(ls, None)
                    clustering(analysis,None)
        '''
        def generic_settings():
            # generic settings
            source = current_dpf_settings['generic']
            dpf.append( 'autodock_parameter_version %s' % 
                        source['autodock_parameter_version'] )
            if not source['parameter_file'] == None:
                parfile = os.path.basename( source['parameter_file'] )
                dpf.append( 'parameter_file %s' %  parfile )
            dpf.append( 'outlev %s' %  source['outlev'] )
            if source['intelec']:
                dpf.append('intelec')
            if current_dpf_settings['search_mode'] == 'sa': seeding = 'pid'
            else: seeding  = 'pid time'
            dpf.append('seed %s' % seeding)
            dpf.append('unbound_model %s' % source['unbound_model'])

        def ligand_settings():
            # XXX UNKNOWN ATOM TYPES ARE NOT HANDLED YET!
            ltypes = sorted(self.LigBook[ligand]['atypes'])
            if 'flex_res' in self.RecBook[receptor].keys():
                ltypes = sorted(list(set(ltypes + self.RecBook[receptor]['flex_atypes'] )))
            dpf.append( 'ligand_types ' + " ".join(ltypes) )
            dpf.append( 'fld %s.maps.fld' % rec_name)
            for a in ltypes:
                line = 'map %s.%s.map' % (rec_name, a)
                dpf.append(line)
            dpf.append( "elecmap %s.e.map" % rec_name )
            dpf.append( "desolvmap %s.d.map" % rec_name )
            dpf.append( 'move %s' % lig_file )
            if self.RecBook[receptor]['is_flexible']: # in self.RecBook[receptor].keys():
                dpf.append( 'flexres %s' % (os.path.basename(self.RecBook[receptor]['flex_res_file'])))
            dpf.append( 'about %2.3f %2.3f %2.3f' % tuple(about) )    

        def ga_settings():
            source =  current_dpf_settings['ga']
            for k in source['KEYWORD_ORDER']:
                dpf.append( "%s %s" % (k, source[k]))

        def sa_settings():
            source =  current_dpf_settings['sa']
            for k in source['KEYWORD_ORDER']:
                dpf.append( "%s %s" % (k, source[k]))

        def ls_settings():
            source =  current_dpf_settings['ls']
            for k in source['KEYWORD_ORDER']:
                if k =='ga_pop_size':
                    if current_dpf_settings['search_mode'] =='ls':
                        dpf.append( "%s %s" % (k, source[k]))
                elif k =='ls_search_freq':
                    if not current_dpf_settings['search_mode'] == 'ls':
                        dpf.append( "%s %s" % (k, source[k]))
                else:
                    dpf.append( "%s %s" % (k, source[k]))

        def set_docking_runs():
            # depending on the self.autodock_search value
            # set the opportune run_command (ga_run, do_local_only)

            # sa : (runs:xx, simanneal)
            # ga : (do_global_only:xx)
            # lga: (ga_run:xx)
            # ls : (do_local_only:xx)
            starting_keyword = { 'lga' : 'ga_run %d',
                                 'ga'  : 'do_global_only %d',
                                 'sa'  : 'runs %d\nsimanneal', # XXX likely to change, check AD 4.3
                                 'ls'  : 'do_local_only %d',
                                 }
            sm = current_dpf_settings['search_mode']
            dpf.append( starting_keyword[sm] % runs )
             
        def clustering_settings():
            source = current_dpf_settings['clustering']
            for k in source['KEYWORD_ORDER']:
                dpf.append("%s %s" % (k, source[k]))

        # starting...
        dpf = []
        if settings == None:
            settings = self.autodock_parameters['dpf_mode']
        if settings == 'template':
            current_dpf_settings = self.autodock_parameters.copy()
        elif settings == 'auto':
            settings = self.dpfComplexityEstimator(lig = lig, 
                rec = rec, grid = self.autodock_parameters['complexity_map'])
        # total runs
        runs = current_dpf_settings['runs']
        # search mode
        search_mode = current_dpf_settings['search_mode']
        # lig properties
        lig_file = os.path.basename( self.LigBook[ligand]['filename'])
        about = getLigAbout( self.LigBook[ligand]['filename'] )
        # rec properties
        rec_file = os.path.basename( self.RecBook[receptor]['filename'] )
        rec_name = os.path.splitext( rec_file )[0]
        generic_settings()
        ligand_settings()
        if search_mode == 'lga':
            ga_settings()
            ls_settings()
        elif search_mode =='ga':
            ga_settings()
        elif search_mode =='sa':
            sa_settings()
        elif search_mode =='ls':
            ls_settings()
        set_docking_runs()
        if current_dpf_settings['do_clustering']:
            clustering_settings()
        return dpf


    def AutoDockMakeGpf(self, receptor, ligand=None, types=None):
        """
            generate GPF for ligand/receptor combination
        """
        if self.debug: print "AutoDockMakeGpf> called with [LIG:%s] [REC:%s] [TYPES:%s]" %  (ligand, receptor, types)
        if ligand == None and types == None:
            if self.debug: print "AutoDockMakeGpf> ERROR no ligand, no types specified"
            return False
        current_gpf = self.autogrid_parameters.copy()

        # ligand atom types
        if ligand:
            # XXX BUG HERE! no unkAtypes recognized!
            ltypes = self.LigBook[ligand]['atypes']
        else:
            ltypes = types
        if self.RecBook[receptor]['is_flexible']:
            #rec_data['flex_atypes'] = flex_res_data['atypes']
            ltypes = list(set(ltypes+ self.RecBook[receptor]['flex_atypes']))
        ltypes = sorted(ltypes)

        # receptor atom types
        rtypes = sorted(self.RecBook[receptor]['atypes'])

        current_gpf['ligand_types']   = " ".join(ltypes)
        current_gpf['receptor_types'] = " ".join(rtypes)
        current_gpf['receptor'] = os.path.basename( self.RecBook[receptor]['filename'] )

        rec_name = os.path.splitext(current_gpf['receptor'])[0]
            
        gpf_file = []
        for k in self.autogrid_parameters['KEYWORD_ORDER']:
        # XXX BUG HERE! missing parameter file code!
            if k == 'gridfld':
                gpf_file.append( "gridfld %s.maps.fld" % (rec_name) )
            elif k == 'gridcenter':
                gpf_file.append( "%s %2.3f %2.3f %2.3f" % ( tuple([k]+current_gpf[k]) ) )
            elif k == 'npts':
                gpf_file.append( "%s %d %d %d" % ( tuple([k]+current_gpf[k]) ) )
            elif k == 'map':
                for a in ltypes:
                    line = 'map %s.%s.map' % (rec_name, a)
                    gpf_file.append(line)
                elec = "elecmap %s.e.map" % rec_name
                desolv = "dsolvmap %s.d.map" % rec_name
                gpf_file.append(elec)
                gpf_file.append(desolv)
            else:
                line = "%s %s" % (k, current_gpf[k])
                gpf_file.append(line)
        return gpf_file
        

    def VinaGenConf(self, ligand=None, receptor=None):
        """
            generate the vina config file
        """
        if self.debug: print "VinaGenConf> called with", ligand, receptor
        vina_settings = self.vina_settings.copy()
        vina_settings['center_x']
        config = []
        if not ligand==None:
            lig = os.path.basename(self.LigBook[ligand]['filename'])
            config.append('ligand = %s' % lig)
        if not receptor == None:
            rec = os.path.basename(self.RecBook[receptor]['filename'])
            config.append('receptor = %s' % rec)
            if self.RecBook[receptor]['is_flexible']:
                flex = os.path.basename(self.RecBook[receptor]['flex_res_file'])
                config.append('flex = %s' % flex)
        for k in vina_settings['KEYWORD_ORDER']:
            if self.debug: print "VinaGenConf> kw[%s] value[%s]" % (k,vina_settings[k])
            if k in vina_settings['OPTIONAL']:
                if vina_settings['OPTIONAL'][k]:
                    config.append('%s = %s' % (k, vina_settings[k]) )
            else:
                config.append('%s = %s' % (k, vina_settings[k]) )
        return config



    def makeLogFunction(self):
        pass

    def TheFunction(self, callback=None):

        # - MAIN_DIR
        #      |
        #      |---REC_1_DIR
        #      |      |
        #      |      |-MAPS [opt,AD-only]
        #      |      |-[ split_suff ]- LIG_1_REC_1_DIR
        #      |      |-[ split_suff ]- LIG_2_REC_1_DIR
        #      |      ...
        #      |
        #      |---REC_2_DIR
        #      |      |
        #      |      |-MAPS [opt,AD-only]
        #      |      |-LIG_1_REC_2_DIR
        #      |      |-LIG_2_REC_2_DIR
        #      |      ...

        # create REC dir
        #   - (A) copy rec file
        #   - (B) calculate grid maps ( + parm file )
        
        # create LIG dir ( + split_number )
        #   - copy lig file
        #   - copy flex file
        #   - make DPF file        ( + parm file)
        #   - (A) make GPF         ( + parm file)
        #   - (B) copy/link maps
        #   - create single_job script

        # create master script
        # 
        # create package
        #
        #
        # Directory naming flow
        # ---------------------
        #
        # current_dir = OUTPUT_DIR
        # 
        #   for r in rec:
        #
        #       rec_dir = OUTPUT_DIR + REC
        #
        #       for l in lig:
        #
        #           lig_dir = rec_dir + SPLIT + LIG
        #

        # get the initial data and freeze static numbers
        ligands = self.ligands()
        receptors = self.receptors()
        lig_types = self.currentLigAtypes()

        self.TOTAL_LIGANDS = len(ligands) # useful for self.makeLigDir!

        # naming section
        cached_maps_dir_name = 'maps'
        gpf_suffix = '_all_maps.gpf'
  
        #self.autogrid_mode = 'always' # 'now' or 'never' (C) 1960 Elvis

        if self.debug: print "TheFunction> start"
        # XXX Numbering voodoo for percentages:
        # tot = rec# * lig# ?
        #       count autogrid step?
        # XXX 
        for rec in receptors: # XXX receptor loop XXX 
            self._recdir = self.makeDir(self.output_dir['path'], rec)
            self._current_rec = rec
            if not self._recdir:
                if self.debug: print "TheFunction> error in creating receptor dir ABORTING"
                return False
            # check map operations
            if self.mode == 'autodock':
                if self.autogrid_mode == 'always':
                    # self._map_dir is used during the calculation
                    # to track the source of the map files
                    # it is populated by:
                    #     - calculateCacheMaps() -> AutGrid run inside Raccoon
                    #     - copyCacheMaps()      -> pre-calc maps are copied from the specified dir
                    self._map_dir = None
                else:
                    self._map_dir = self.makeDir(self._recdir, cached_maps_dir_name)
                    if self.autogrid_mode == 'now':     # Raccoon maps
                        self.calculateCacheMaps(mapdir=self._map_dir, lig_types=lig_types)
                    elif self.autogrid_mode == 'never': # pre-cached maps
                        if self.debug: print self._map_dir, lig_types
                        self.copyCacheMaps(target=self._map_dir, lig_types=lig_types, file_policy ='copy')
            self.LIGAND_COUNTER = 0
            for lig in ligands: # 
                self._current_lig = lig # XXX possibly useless
                self._curr_lig_dir = self.makeLigDir(lig)
                # copy receptor rigid+flex
                self.copyRecFiles(rec, self._curr_lig_dir)
                # copy ligand file and flex if necessary
                self.copyLigFiles(lig, self._curr_lig_dir)
                if self.mode == 'autodock':
                    #   provide map data (GPF/cached)
                    if self.autogrid_mode == 'always':
                        self.prepareLigGpf(path=self._curr_lig_dir, rec=rec, lig=lig)
                    else:
                        self.copyCacheMaps(self._map_dir, lig_types, self.autogrid_map_file)
                    # generate DPF
                    self.prepareLigDpf(path=self._curr_lig_dir, rec=rec, lig=lig)
                    # copy Parmfile if necessary
                    # XXX self.copyAdParmFile()
                elif self.mode == 'vina':
                    self.prepareLigVina(path=self._curr_lig_dir, rec=rec, lig=lig)
                # update the jobs list status...
                # XXX store the lig_dir for the job script
                self.LIGAND_COUNTER += 1

            # XXX update the master-job list status 
            # and generate the actual script here


    def prepareLigVina(self, path, rec, lig):
        """ generate GPF for the lig/rec combination"""
        conf = self.VinaGenConf(ligand=lig, receptor=rec)
        conf_filename = path + os.sep + ('%s_%s.conf' % (lig, rec) )
        writeList( conf_filename, conf, addNewLine=True)
        return conf_filename
        

    #def provideMapData(self, lig, rec=None):
    #    """handle the grid data issue,
    #    """
    #    self._current_rec # curent receptor that is going processed
    #    self._current_lig # current lig processed

    def calculateCacheMaps(self, mapdir, rec, lig_types):
        """ run AutoGrid n the to calculate grid maps on the atom 
            types requested for chaching
        """
        # prepare the AutoGrid calculation 
        gpf = self.prepareCacheMapCalc(path=mapdir, rec=rec, lig_types=lig_types)
        # run AutoGrid at the specified path
        self.runAutoGrid(path, gpf)
        # self.autogrid_maps = {} # { 'A': 'protein.A.map', 'C' : 'protein.C.map', 
        return True


    def prepareCacheMapCalc(self, path, rec, lig_types): 
        """ generate the calculation for caching map files
            - copy receptor files ( + flex )
            - generate GPF for all requested ligand types
        """
        # copy receptor files
        response = self.copyRecFiles( path, rec )
        # XXX if parameter files specified copy it here

        self.copyAdParmFile()
        if response['success']:
            # generate the gpf
            gpf = self.AutoDockMakeGpf(receptor = rec, types=lig_types)
            gpf_filename = self._mapdir + os.sep + rec + gpf_suffix
            writeList( gpf_filename, gpf, addNewLine=True)
            return gpf_filename
        else:
            return False

    def prepareLigGpf(self, path, rec, lig):
        """ generate GPF for the lig/rec combination"""
        gpf = self.AutoDockMakeGpf(receptor = rec, ligand = lig)
        gpf_filename = path + os.sep + rec + '.gpf'
        writeList( gpf_filename, gpf, addNewLine=True)
        return gpf_filename
        
    
    def prepareLigDpf(self,path, rec, lig, settings=None):
        """generate DPF for the lig/rec combination

            'settings' is an option to specify if the DPF will
            be generated:

              - 'template' : by using the same settings template 
                             for every ligand

              - 'auto'     : by adapting the search parameters
                             to the bindig volume and the ligand
        """
        if settings == None and self.autodock_parameters['dpf_mode'] == None:  
            settings = None
        elif settings == 'auto':
            settings = self.dpfComplexityEstimator(lig = lig, 
                rec = rec, grid = self.autodock_parameters['complexity_map'])

        dpf = self.AutoDockMakeDpf(receptor=rec, ligand=lig, settings = None)

        dpf_filename = path + os.sep + "%s_%s"  + '.dpf'
        dpf_filename = dpf_filename % (lig, rec)
        writeList( dpf_filename, dpf, addNewLine=True)
        return dpf_filename
        

    def dpfComplexityEstimator(self, lig, rec, grid=None):
        """ estimate the complexity of the search by analyzing
            the grid box volume and ligand properties
            (+ flex_res, if defined )
        """
        if grid == None:
            grid = self.autodock_parameters['complexity_map'] 
        # XXX not working yet...
        return self.autodock_parameters.copy()
        


    def copyAdParmFile(self, path, filename):
        """copy parameter AD parameter file to requested path
        """
        # XXX it doesn't work yet!
        print "copyAdParmFile> checking if parm.dat requested [NOT WORKING YET]"
        pass

    def runAutoGrid(self, path, gpf, binary=None):
        """run AutoGrid binary in the specified path using
            the GPF specified.
            By default the self.autogrid_bin is used
        """
        # XXX remember to trigger the update of the  self.autogrid_options['complexity_map']
        #     map used to calculate ligand accessibility (usually *.C.map)
        if binary == None:
            binary = self.autogrid_bin
        # cd to path
        # extract basename(gpf)
        # check if gpf is present
        # run autogrid

    def copyCacheMaps(self, target, lig_types = [], file_policy = 'copy'):
        """ copy cache map files in place
            'mode' can be 'copy' or 'link'
            
            this function can be used to copy map files in 
            ligand directories or on cached_map directories
        """
        # self.autogrid_maps = {} # { 'A': 'protein.A.map', 'C' : 'protein.C.map', 
        if file_policy == 'copy':
            copy_func = shutil.copy2
        elif file_policy == 'link':
            copy_func = os.symlink
        for a in lig_types :
            copy_func( self.autogrid_maps[a], target)



    def copyRecFiles(self, rec, target_dir):
        """ copy receptor file and flex_res, if specified
            in the directory specified
        """
        try: 
            shutil.copy2( self.RecBook[rec]['filename'], target_dir)
            reason = 'rigid'
            if self.RecBook[rec]['is_flexible']:
                shutil.copy2( self.RecBook[rec]['flex_res_file'], target_dir)
                reason += '_flex'
            return { 'success': True, 'reason': reason }
        except:
            error = "%s" % exc_info()[1]
            return { 'success': False, 'reason': error }

    def copyLigFiles(self,lig, target_dir):
        """ copy ligand file and flex_res, if specified,
            in the directory specified
        """
        try: 
            shutil.copy2( self.LigBook[lig]['filename'], target_dir)
            reason = 'ligand'
            #if self.RecBook[self._current_rec]['is_flexible']:
            #    shutil.copy2( self.RecBook[self._current_rec]['flex_res_file'], target_dir)
            #    reason += '_flex'
            return { 'success': True, 'reason': reason }
        except:
            error = "%s" % exc_info()[1]
            return { 'success': False, 'reason': error }

        


    def makeDir(self, path, name):
        """ create dir for the generation process"""
        dirname = path + os.sep + name
        try:
            os.makedirs(dirname)
            return dirname
        except:
            print "makeDir> ERROR creating dir [%s]: %s" % (dirname, exc_info()[1])
            return False



    def makeLigDir(self, ligand):
        """create the ligand directory, adding 
           the splitting suffix, if necessary
        """
        if not self.generalOpts['generic']['split'] == None:
            split = self.generalOpts['generic']['split']
            if self.TOTAL_LIGANDS <= split:
                suffix = ''
            else:
                zeropadding = "%%0%dd" % len(str(self.LIGAND_COUNTER/split))
                suffix = (zeropadding+os.sep) % (self.LIGAND_COUNTER/split)
        else: suffix = ''
        dir_name = self._recdir+os.sep+suffix+ligand+os.sep
        try:
            os.makedirs(dir_name)
        except:
            if self.debug: print "problem: %s" % exc_info()[1]
            return False    
        return dir_name


if __name__ == '__main__':

    def usage():
        print "Usage: raccoon -d . -r protein.pdbqt -x"
    lig_list = ['data/lig/ZINC00913760.pdbqt',
                'data/lig/ZINC01365912.pdbqt']

    rec_list = ['data/rec/single/frame_001.pdbqt',
                'data/rec/single/frame_002.pdbqt',
                'data/rec/flex/1JFF_rigid.pdbqt',
                'data/rec/flex/1JFF_flex.pdbqt',
                ]

    xray_lig = 'data/lig/ZINC00913760.pdbqt'

    autoligand = 'data/al/FILL_250out1.pdb'


    gpf_template = 'data/etc/template_without_paramfile.gpf'
    conf_template = 'data/etc/test.conf'
    map_policy = '' # always, now, never

    gpf_center = [ 0.5, -12, 24 ]
    gpf_size   = [ 15.5, 20, 10 ]
    gpf_center_WRONG = [1,2,'a']
    gpf_size_WRONG = [1,2,'a']


    dpf_template = None
    flex_res_names = []
    flex_res_file = ''

    out_dir = 'data/NEW_CLASS_TEST/aaa'
    #out_dir = '/media/iguana/TEST_RACCOON_SPEED/'

    #lig_list = getLines('/media/mammoth/database/zinc/ChEBI/ChEBI_total.2011.8.3/files_pdbqt.lst', doStrip=True)

    test = 'autodock'
    #test = 'vina'


    ##################################################################
    # TESTING 

    raccoon = Raccoon(debug=True)
    # adding a ligand list
    raccoon.setMode('vina')

    print "- importing ligands... ",
    raccoon.addLigandList(lig_list)
    print len(raccoon.ligands())

    # adding a receptor list
    raccoon.addReceptorList(rec_list)
    print "- check receptor status"
    for r in raccoon.RecBook.keys():
        rec = raccoon.RecBook[r]
        if 'flex_res' in rec.keys():
            flex = " %s : %s types(%s)" % (",".join(rec['flex_res']),
                 rec['flex_res_file'], ",".join(rec['flex_atypes'])) 
        else:
            flex = 'None'
        print "\t%s : rigid[%s] \t flex[%s]" % (r,
            raccoon.RecBook[r]['filename'], flex)

    # setting the grid center (tuple, vector...)
    #raccoon.setBoxCenter([1,2,'a'])
    raccoon.setBoxCenter(gpf_center)

    # setting the grid size (tuple, vector...)
    #raccoon.setBoxSize([1,2,'a'])
    raccoon.setBoxSize(gpf_size)

    # setting the output dir
    response = raccoon.setOutputDir(out_dir)
    print "\n- check selected directory properties ( %s : %s ):" % response
    for i in raccoon.output_dir.keys():
        print "\t%s : %s" % (i, raccoon.output_dir[i])
    #self.output_dir['freespace'] = checkDiskSpace(dir_path)


    # testing filtering sets

    # setting custom filter settings
    custom_filter_set = raccoon.filterSettings.copy()
    custom_filter_set['hba'][0] = 28
    custom_filter_set['hbd'][1] = 76  
    custom_filter_set['title'] = 'TESTING'
    raccoon.setFilterSet(customFilter=custom_filter_set)
    print "\n- current filter set", raccoon.filterSettings

    # perform the filter with the current settings
    raccoon.filterLigands()
    print "- previewing ligands passing current filter:",
    print raccoon.filterLigands(previewOnly=True)

    raccoon.filterLigands()
    print "- filtering with '%s'" % (raccoon.filterSettings['title'])
    print "- updating status..."
    status = raccoon.updateStatus(report=True)


    # setting default filter settings (lipinsky)
    raccoon.setFilterSet(filterSet = 'lipinsky')
    print "\n- current filter set ", raccoon.filterSettings
    
    # perform the filter with the current settings
    raccoon.filterLigands()
    print "- filtering with '%s'" % (raccoon.filterSettings['title'])
    print "- updating status..."
    status = raccoon.updateStatus(report=True)


    gpf_source = 'lig' # gpf, lig, al, conf
    gpf_source = 'conf'

    if gpf_source == 'gpf':
        # use template GPF to set center and size"
        print "\n- setting the center with GPF (%s) " % gpf_template
        # raccoon.gridBoxFromGpf(gpf_template)
        raccoon.importGpfTemplate(gpf_template)

        print "\tcenter (%2.3f, %2.3f, %2.3f)" % (raccoon.box_center[0],
                    raccoon.box_center[1], raccoon.box_center[2] )
        print "\tsize   (%2.3f, %2.3f, %2.3f)" % (raccoon.box_size[0],
                    raccoon.box_size[1], raccoon.box_size[2] )

    elif gpf_source == 'lig':
        # use reference ligand to set center and size"
        print "\n- setting center with ligand (%s)" % xray_lig
        raccoon.gridBoxFromLigand(xray_lig, tolerance=5)

        print "\tcenter (%02.3f, %02.3f, %02.3f)" % (raccoon.box_center[0],
                    raccoon.box_center[1], raccoon.box_center[2] )
        print "\tsize   (%02.3f, %02.3f, %02.3f)" % (raccoon.box_size[0],
                    raccoon.box_size[1], raccoon.box_size[2] ) 

    elif gpf_source == 'al':
        # use AutoLigand result to set center and size"
        print "\n- setting center with ligand (%s)" % autoligand
        raccoon.gridBoxFromLigand(autoligand, tolerance=2)

        print "\tcenter (%02.3f, %02.3f, %02.3f)" % (raccoon.box_center[0],
                    raccoon.box_center[1], raccoon.box_center[2] )
        print "\tsize   (%02.3f, %02.3f, %02.3f)" % (raccoon.box_size[0],
                    raccoon.box_size[1], raccoon.box_size[2] ) 

    elif gpf_source == 'conf':
        print '\n- setting center with config file (%s)' % conf_template

        raccoon.gridBoxFromConf(conf_template)
        print "\tcenter (%02.3f, %02.3f, %02.3f)" % (raccoon.box_center[0],
                    raccoon.box_center[1], raccoon.box_center[2] )
        print "\tsize   (%02.3f, %02.3f, %02.3f)" % (raccoon.box_size[0],
                    raccoon.box_size[1], raccoon.box_size[2] ) 
        


    # generate GPF for ligands/rec combinations
    raccoon.autodock_parameters['search_mode'] = 'lga'

    if test == 'autodock':
        raccoon.setMode('autodock')
    elif test =='vina':
        raccoon.setMode('vina')

    print 
    raccoon.TheFunction()
    exit()



    """
    for r in raccoon.receptors():
        print "\n##########\nREC [ %s ]" % ( raccoon.RecBook[r]['name'] )
        for l in raccoon.ligands():
            print "\t - GPF for %s" % l
            settings = { 'center' : raccoon.box_center,
                         'size'   : raccoon.box_size,
                       }
            box = raccoon.gridBox()
            gpf = raccoon.AutoDockMakeGpf(receptor=r, ligand=l)
            outname = "%s_%s.gpf" % (l, r)
            outname = raccoon.output_dir['path']+os.sep+outname
            writeList(outname, gpf, addNewLine=True)
    
            dpf=raccoon.AutoDockMakeDpf(ligand=l, receptor=r)
            outname = "%s_%s.dpf" % (l, r)
            outname = raccoon.output_dir['path']+os.sep+outname
            writeList(outname, dpf, addNewLine=True)

            conf=raccoon.VinaGenConf(ligand=l, receptor=r)
            outname = "%s_%s.conf" % (l, r)
            outname = raccoon.output_dir['path']+os.sep+outname
            writeList(outname, conf, addNewLine=True)
            

    if test == 'autodock':
        # gpf
        pass

        # dpf

    elif test =='vina':
        raccoon.setMode(mode ='vina')
    """





# NOTES
# - leading zeroes for Mol2 http://docs.python.org/release/2.4.4/lib/typesseq-strings.html
# - string.rjust(value, 'padder')
# YELD for ITERATORS: http://docs.python.org/tutorial/classes.html



#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

from mglutil.events import Event, EventHandler

class RaccoonEventManager(EventHandler):
    def __init__(self):
        EventHandler.__init__(self)


class SetResourceEvent(Event):
    """ class handle the resource definition/change
            [local, cluster, opal] """

    def __init__(self, resource):
        Event.__init__(self)
        self.resource = resource


class SwitchConfigCamera(Event):
    """ handle activation/deactivation of the grid box config
        camera (with preventive disactivation of other cameras?)
    """

    def __init__(self, app, is_active, *arg, **kw): # XXX remove APP
        Event.__init__(self, *arg, **kw) #
        if is_active:
            # deactivate other cameras
            # should this class be an event dispatcher?
            pass
        print "SwitchConfigCamera> [%s -> %s] "% (self.app.isConfigCameraActive, is_active)

        self.app.isConfigCameraActive = is_active


class SetDockingEngine(Event):
    """ handle selection of docking engine (AD/Vina)
    """
    def __init__(self, engine, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        self.dockingengine = engine


class SetDefaultDataPath(Event):
    """ set the default data path (i.e. from preferences)
    """
    def __init__(self, path, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        self.path = path



class SshServerManagerEvent(Event):
    """ triggered when a server is added/removed
        by the ServerManager

    """
    def __init__(self,  *arg, **kw):
        Event.__init__(self, *arg, **kw)

class ServerConnection(Event):
    """ event triggered when a connection
        is established with a server
    """
    def __init__(self,  *arg, **kw):
        Event.__init__(self, *arg, **kw)
        self.is_connected = True



class ServerDisconnection(Event):
    """ event triggered when a connection
        with a server is closed/broken
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        self.is_connected = False


class UpdateServerListEvent(Event):
    """ event triggered by a change in the list
        of known servers
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        #self.is_connected = False


class ServerFeaturesRefresh(Event):
    """ triggered when new features are added to the server
        (i.e. new libraries uploaded, new services installed
        via the GUI or by the user)
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)

class LibrariesUpdate(Event):
    """ triggered when the number of libraries on the server
        changes.
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)


class ServicesUpdate(Event):
    """ triggered when services on the server
        get installed or deleted.
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)

class ServiceSelected(Event): 
    """ triggered when selecing a service
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)


###### RECEPTOR



class ReceptorListChange(Event):
    """ triggered when new features are added to the server
        (i.e. new libraries uploaded, new services installed
        via the GUI or by the user)
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)


##### CONFIG EVENTS



class SearchConfigChange(Event):
    """ triggered when there's a change in the search configuration
        (i.e. box definition, search parm...)
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        

#### JOBS EVENTS


class UpdateJobHistory(Event):
    """ triggered when a job is submitted, or a new job history file
        is read
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        # this should be changed to be a list? prj->exp->vs
        #self.obj = kw.get('obj', None)
        self.prj = kw.get('prj', None)
        self.exp = kw.get('exp', None)
        self.name = kw.get('name', None)
        self.jtype = kw.get('jtype', None)
        self.properties = kw.get('properties', None)
        #self.status = kb['status']


class DeletedJobHystoryItem(Event):
    """ triggered when a hystory item is deleted
        the event can be defined as one of the following:
            prj, exp, vs   <-- delete vs
            prj, exp       <-- delete exp
            prj            <-- delete prj
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        # XXX see if it must be a multiple object type or only VS
        self.prj = kw.get('prj', None)
        self.exp = kw.get('exp', None)
        self.name = kw.get('name', None)
        



class SyncJobHistory(Event):
    """ triggered when connecting to a server to
        import new unknown jobs
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        
        
class UserInputRequirementUpdate(Event):
    """ triggered when user set/unset one of the requirements
        for job submission
        _type = 'ligand', 'receptor', 'config', ...
         ...???
    """
    def __init__(self, _type=None, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        self._type = _type
      

##### ANALYSIS EVENTS

class FilterSetSelection(Event):
    """ triggered when a filter set is selected
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        
class ResultsImportedDeleted(Event):
    """ triggered when results are added or removed
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)

class FilterRunEvent(Event):
    """ triggered when a filtering process is performed
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)


class FilterInteractionEvent(Event):
    """ triggered when a filtering process is performed
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)






class LigandResultPicked(Event):
    """ triggered when a filtering process is performed
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)
        #print "LIGPICK", arg, kw
        self.ligname = kw['ligname']
        self.jobname = kw['jobname']
        self.pose = kw['pose']

class LigandResultNuked(Event):
    """ triggered when all ligand results are nuked
    """
    def __init__(self, *arg, **kw):
        Event.__init__(self, *arg, **kw)



#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import CADD.Raccoon2
import RaccoonBasics as rb
import RaccoonEvents
import RaccoonServers
import RaccoonServices
import CADD.Raccoon2.HelperFunctionsN3P as hf
import RaccoonProjManTree
import os, Pmw
from PmwOptionMenu import OptionMenu as OptionMenuFix
import Tkinter as tk
import tkMessageBox as tmb
import tkFileDialog as tfd
from PIL import Image, ImageTk
# mgl modules
from mglutil.events import Event, EventHandler
from mglutil.util.callback import CallbackFunction # as cb

#import EF_resultprocessor 



class JobManagerTab(rb.TabBase, rb.RaccoonDefaultWidget):
    """ populate and manage the job manager tab """
    
    def __init__(self, app, parent, debug=False):
        # get
        rb.TabBase.__init__(self, app, debug)
        rb.RaccoonDefaultWidget.__init__(self, parent)
        self.initIcons()
        self.resource = self.app.resource
        # Events
        self.app.eventManager.registerListener(RaccoonEvents.SetResourceEvent, self.handleResource) # set resource
        self.app.eventManager.registerListener(RaccoonEvents.ServerConnection, self._updateRequirementsSsh) # open connection
        #self.app.eventManager.registerListener(RaccoonEvents.UpdateJobHistory, self.updateJobTree)  # job history update
        self.app.eventManager.registerListener(RaccoonEvents.ServiceSelected, self.updateRequirements) # docking service is selected 
        self.app.eventManager.registerListener(RaccoonEvents.UserInputRequirementUpdate, self.updateRequirements) # data input (lig,rec...)
        self.app.eventManager.registerListener(RaccoonEvents.SearchConfigChange, self.updateRequirements) # search config change (box)
        self._buildjobman(self.parent)

    def _buildjobman(self, target):
        """ build the job manager tree"""
        pgroup = Pmw.Group(target, tag_text = 'Jobs', tag_font=self.FONTbold)
        #tk.Button(pgroup.interior(), text='Refresh', image='self.'
        self.jobtree = RaccoonProjManTree.VSresultTree(pgroup.interior(), app = self.app, iconpath=self.iconpath)
        pgroup.pack(expand=1, fill='both', anchor='n', side='bottom')
        self.initJobTree()


    def initJobTree(self, event=None):
        """ populate the tree with the history filer"""
        self.jobtree.setDataFile(self.app.getHistoryFile())

    def initIcons(self):
        """ initialize the icons for the interface"""
        self.iconpath = icon_path = CADD.Raccoon2.ICONPATH 
        f = icon_path + os.sep + 'system.png'
        self._ICON_sys = ImageTk.PhotoImage(Image.open(f))
        f = icon_path + os.sep + 'submit.png'
        self._ICON_submit = ImageTk.PhotoImage(Image.open(f))




    def handleResource(self, event=None):
        self.setResource(event.resource)


    def setResource(self, resource):
        '''adapt the job manager panel to reflect currently selected resource'''
        if resource == 'local':
            self.setLocalResource()
        elif resource == 'cluster':
            self.setClusterResource()
        elif resource == 'opal':
            self.setOpalResource()

    def setLocalResource(self):
        #self.frame.pack_forget()
        self.resetFrame()
        #self.frame = tk.Frame(self.group.interior())

        tk.Label(self.frame, text='(local) requirement widget 1').pack()
        tk.Label(self.frame, text='(local) requirement widget 2').pack()
        tk.Label(self.frame, text='(local) requirement widget 3').pack()
        tk.Label(self.frame, text='SUBMIT').pack()
        self.frame.pack(expand=1, fill='both')


        #print "Raccoon GUI job manager is now on :", self.app.resource

    def setClusterResource(self):
        self.resetFrame()

        #self.frame.configure(bg='red')
        self.group = Pmw.Group(self.frame, tag_text = 'Cluster submission requirements', tag_font=self.FONTbold)
        f = self.group.interior()
        #f.configure(bg='red')

        lwidth = 20
        rwidth = 60
        lbg = '#ffffff'
        rbg = '#ff8888'
        fg = 'black'
        # server connection
        tk.Label(f, text='Server', width=lwidth, font=self.FONT,anchor='e').grid(row=1, column=1,sticky='e',padx=5, pady=0)
        self.reqConn = tk.Label(f, text = '[ click to connect ]', fg = fg, bg=rbg, width=rwidth, font=self.FONT, **self.BORDER)
        self.reqConn.grid(row=1,column=2, sticky='e')
        cb = CallbackFunction(self.switchtab, 'Setup')
        self.reqConn.bind('<Button-1>', cb)
        
        # XXX self.GUI_LigStatus.bind('<Button-1>', lambda x : self.notebook.selectpage('Ligands')) 

        # docking service
        # ligands
        tk.Label(f, text='Docking service', width=lwidth, font=self.FONT,anchor='e').grid(row=2, column=1,sticky='e',padx=5, pady=1)
        self.reqService = tk.Label(f, text = '[ click to select ]', fg = fg, bg=rbg, width=rwidth, font=self.FONT, **self.BORDER)
        self.reqService.grid(row=2,column=2, sticky='w', pady=1)
        cb = CallbackFunction(self.switchtab, 'Setup')
        self.reqService.bind('<Button-1>', cb)


        # ligands
        tk.Label(f, text='Ligands', width=lwidth, font=self.FONT,anchor='e').grid(row=3, column=1,sticky='e',padx=5, pady=1)
        self.reqLig = tk.Label(f, text = '[ click to select ]', fg = fg, bg=rbg, width=rwidth, font=self.FONT, **self.BORDER)
        self.reqLig.grid(row=3,column=2, sticky='w', pady=1)
        cb = CallbackFunction(self.switchtab, 'Ligands')
        self.reqLig.bind('<Button-1>', cb)

        # receptor
        tk.Label(f, text='Receptors', width=lwidth, font=self.FONT,anchor='e').grid(row=5, column=1,sticky='e',padx=5, pady=0)
        self.reqRec = tk.Label(f, text = '[ click to select ]', fg = fg, bg=rbg, width=rwidth, font=self.FONT, **self.BORDER)
        self.reqRec.grid(row=5,column=2, sticky='w')
        cb = CallbackFunction(self.switchtab, 'Receptors')
        self.reqRec.bind('<Button-1>', cb)

        # config
        tk.Label(f, text='Config', width=lwidth, font=self.FONT,anchor='e').grid(row=7, column=1,sticky='e',padx=5, pady=1)
        self.reqConf = tk.Label(f, text = '[ click to select ]', fg = fg, bg=rbg, width=rwidth, font=self.FONT, **self.BORDER)
        self.reqConf.grid(row=7,column=2, sticky='w', pady=1)
        cb = CallbackFunction(self.switchtab, 'Config')
        self.reqConf.bind('<Button-1>', cb)

        # scheduler
        #tk.Label(f, text='Scheduler', width=lwidth, font=self.FONT,anchor='e').grid(row=9, column=1,sticky='e',padx=5, pady=0)
        #self.reqSched = tk.Label(f, text = '[ click to select ]', fg = fg, bg=rbg, width=rwidth, font=self.FONT, **self.BORDER)
        #self.reqSched.grid(row=9,column=2, sticky='w')
        #cb = CallbackFunction(self.switchtab, 'Setup')
        #self.reqSched.bind('<Button-1>', cb)

        # submission
        self.SubmitButton = tk.Button(f, text = 'Submit...', image=self._ICON_submit, 
            font=self.FONT, compound='left',state='disabled', command=self.submit, **self.BORDER)
        self.SubmitButton.grid(row=20, column=1, sticky='we', columnspan=3, padx=4, pady=3)

        self.group.pack(fill='none',side='top', anchor='w', ipadx=5, ipady=5)

        self.frame.pack(expand=0, fill='x',anchor='n')
        self._updateRequirementsSsh()

        #print "Raccoon GUI job manager is now on :", self.app.resource


    def switchtab(self, tab, event=None):
        """ """
        self.app.notebook.selectpage(tab)

    def setOpalResource(self):
        self.resetFrame()
        self.frame = tk.Frame(self.group.interior())
        tk.Label(self.frame, text='(opal) requirement widget 1').pack()
        tk.Label(self.frame, text='(opal) requirement widget 2').pack()
        tk.Label(self.frame, text='(opal) requirement widget 3').pack()
        self.frame.pack(expand=1, fill='both')
        #print "Raccoon GUI job manager is now on :", self.app.resource

    def updateRequirements(self, event=None):
        """ update submission requirements """
        if self.app.resource == 'local':
            self._updateRequirementsLocal(event)
        elif self.app.resource == 'cluster':
            self._updateRequirementsSsh(event)
        elif self.app.resource == 'opal':
            self._updateRequirementsOpal(event)
    
    def submit(self, event=None, suggest={}):
        """ find out which EVENT should be triggered and how"""
        jsub = JobSubmissionInterface(self.frame, jmanager=self.jobtree, app = self.app, suggest=suggest)
        job_info = jsub.getinfo()
        self.app.setBusy()
        if job_info == None:
            self.app.setReady()
            return
        if self.app.resource == 'local':
            self.submit_local(job_info)
        elif self.app.resource == 'cluster':
            self.submit_cluster(job_info)
        elif self.app.resource == 'opal':
            self.submit_opal(job_info)
        self.app.setReady()

    def submit_local(self, job_info):
        """ manage submission and feedback from local resource"""
        report = self.app.submitLocal(job_info)

    def submit_opal(self, job_info):
        """ manage submission and feedback from Opal resource"""
        report = self.app.submitOpal(job_info)

    def submit_cluster(self, job_info):
        """ manage submission and feedback from Ssh cluster resource"""
        # report = { 'submissions': [], 'server_duplicates' : [], 'local_duplicates' : [] }
        report = self.app.testSshJobs(job_info)
        sdup = report['server_duplicates']
        ldup = report['local_duplicates']
        choice = 'skip'
        #print "\n\n\nREPORT FOR DUPLICATES", report
        #print "\n\n\n"
        if len(sdup) or len(ldup):
            #print "WE HAVE DUPLICATES>", len(sdup), len(ldup)
            choice = ManageJobOverlaps(self.frame, report)
            #buttons = ('Skip', 'Modify tag', 'Auto-rename', 'Overwrite', 'Cancel'),
            if choice == 'tag':
                t = 'Submission'
                i = 'info'
                m = 'Repeat the submission specifying a different tag.'
                tmb.showinfo(parent=self.frame, title=t, message=m, icon=i)
                return
        # XXX close here the job submission manager
        if choice == 'cancel':
            return

        self.app.setBusy()
        submission = self.app.submitSsh(job_info, duplicates=choice)
        s = len(submission)
        if s>0:
            t = 'Submission'
            i = 'info'
            m = '%d jobs submitted successfully.' % (s / 2)
            tmb.showinfo(parent=self.frame, title=t, message=m, icon=i)
            self.app.setReady()
            return True
        else:
            t = 'Submission'
            i = 'error'
            m = 'No jobs have been submitted!'
            tmb.showinfo(parent=self.frame, title=t, message=m, icon=i)
            self.app.setReady()
            return False

    def _updateRequirementsLocal(self, event=None):
        """ update the check for requirements 
            of local submission
        """
        _type = event._type
        pass


    def _updateRequirementsSsh(self, event=None):
        """ update the check for requirements 
            of ssh submission
        """
        g = 'black'
        r = 'red'
        d = '[ click to select ]'
        green = '#99ff44'
        red = '#ff4444'
        orange = '#ffcc44'

        missing = False
        #_type = event._type
        # check connection
        if not self.app.server == None:
            t = "connected to %s" % self.app.server.properties['name']
            # check racconized
            if self.app.server.properties['ready']:
                self.reqConn.configure(text = t, bg = green)
            else:
                t = "connected to %s (NOT RACCONIZED!)" % self.app.server.properties['name']
                self.reqConn.configure(text = t, bg = orange)
                missing = True
        else:
            missing = True
            self.reqConn.configure(text = d, bg = red)

        if not self.app.dockingservice == None:
            t = '%s' % self.app.dockingservice
            self.reqService.configure(text = t, bg = green)
        else:
            missing = True
            self.reqService.configure(text = d, bg = red)
        
        # ligand 
        if len(self.app.ligand_source):
            libnames = ",".join([x['lib'].name() for x in self.app.ligand_source])
            t = "library selected (%s)" % libnames
            self.reqLig.configure(fg = g, text = t, bg = green)
        else:
            missing = True
            self.reqLig.configure(fg = g, text = d, bg = red)
        # rec
        if len(self.app.engine.receptors()) > 0:
            t = "%s receptors selected" % len(self.app.engine.receptors())
            self.reqRec.configure(fg = g, text = t, bg = green)
        else:
            missing = True
            self.reqRec.configure(fg = g, text = d, bg = red)
        # conf
        conf = self.app.engine.gridBox()
        if not None in conf['center'] + conf['size']:
            t = "search box defined"
            self.reqConf.configure(fg = g, text = t, bg = green)
            c = g
        else:
            missing = True
            self.reqConf.configure(fg = g, text = d, bg = red)
        # sched
        #if not self.app.server == None:
        #    sched = self.app.server.systemInfo('scheduler')
        #else:
        #    missing = True
        # submit
        if not missing:
            self.SubmitButton.configure(state='normal')

    def _updateRequirementsOpal(self, event=None):
        """ update the check for requirements 
            of opal submission
        """
        _type = event._type
        pass

class JobSubmissionInterface(rb.RaccoonDefaultWidget):
    """ ask for Project, Exp, VS info..."""

    def __init__(self, parent, jmanager, app, suggest={}):
        """ parent      : tkparent
            jmanager    : job manager tree (to query current prj,exp...)
            app         : containing app
        """
        rb.RaccoonDefaultWidget.__init__(self, parent)
        self.jmanager = jmanager
        self.app = app
        self._new = '<new>'
        self.jobdata = None
        self.suggest = suggest
        self.initIcons()
        self.build()

    def initIcons(self):
        pass

    def close(self, result):
        """ close the window and decides what to do
            if OK requested, check values and start submission
        """
        if result == 'OK':
            if not self.checkinfo():
                return
            p = self.getPrj()
            e = self.getExp()
            t = self.tag_entry.getvalue().strip()
            self.jobdata = {'prj' : p, 'exp': e, 'tag':t} 
            self.win.deactivate(self.jobdata)
        else:
            self.win.deactivate(False)

    def getPrj(self):
        """ return the project name"""
        m = ('The project name is not valid.\n')
        p = self.prj_pull.getvalue() # old project
        if p == self._new:
            if not self.prj_new.valid():
                self.errorMsg(m)
                return False
            p = self.prj_new.getvalue()
        return p

    def getExp(self):
        """ return the experiment name"""
        m = ('The experiment name is not valid.\n')
        e = self.exp_pull.getvalue() # old project
        if e == self._new:
            if not self.exp_new.valid():
                self.errorMsg(m)
                return False
            e = self.exp_new.getvalue()
        return e

    def getTag(self):
        """ return the tag"""
        tag = self.tag_entry.getvalue()
        return tag


    def checkDuplicates(self):
        """ check that the jobs that are going to be 
            submitted do not have the same name of 
            already submitted jobs
        """
        m = ('The submission cannot be performed because there '
             'are already jobs with the same name stored in project %s '
             '/experiment %s.\n\n'
             'Either create new project/experiments or use a different tag.')
        job_info = {'prj' : self.getPrj(),
                    'exp' : self.getExp(),
                    'tag' : self.getTag()
                   }
        report = self.app.testSshJobs(job_info)
        e = []
        if len(report['server_duplicates']):
            e.append('the current server')
        if len(report['local_duplicates']):
            e.append('the local client')
        if len(e):
            m = m % ( job_info['prj'], job_info['exp'])
            self.errorMsg(m)
            return False
        return True

    def checkinfo(self):
        """ check that user provided info are valid"""
        if not self.getPrj():
            return False
        if not self.getExp():
            return False
        if not self.checkDuplicates():
            return False
        return True


    def errorMsg(self, message):
        """ display submission entries error"""
        t = 'Incorrect name entry'
        i = 'error'
        tmb.showinfo(parent=self.win.interior(), title=t, message=message, icon=i)
        return


    def getinfo(self):
        """ return the user provided info"""
        return self.jobdata


    def _setprjname(self, event=None):
        choice = self.prj_pull.getvalue()
        if choice == self._new:
            self.prj_new.grid(row=4, column=2, sticky='we',padx=4,pady=4)
            self.prj_new.checkentry()
        else:
            self.prj_new.grid_forget()
        exp_list = self._getexplist()
        self.exp_pull.setitems( exp_list )
        self.exp_pull.setvalue( exp_list[-1])
        self.exp_pull.invoke()

    def _getexplist(self, event=None):
        prj = self.prj_pull.getvalue()
        if prj == self._new:
            return [self._new]
        else:
            exp_list = sorted(self.info[prj].keys())
            return exp_list + [self._new]

    def _setexpname(self, event=None):
        choice = self.exp_pull.getvalue()
        if choice == self._new:
            self.exp_new.grid(row=8,column=2, sticky='we', padx=4,pady=4)
            self.exp_new.checkentry()
        else:
            self.exp_new.grid_forget()

    def build(self):
        # get info from the current manager
        self.info = self.jmanager.getTreeGraph()
        self.prj_list = sorted(self.info.keys()) + [self._new]
        #self.prj_list.append(self._new)
        self.win = Pmw.Dialog(parent=self.parent, buttons=('OK', 'Cancel'),
            title = 'Submit jobs', command = self.close)
        w = self.win.interior()
        bbox = self.win.component('buttonbox')
        for i in range(bbox.numbuttons()):
            bbox.button(i).configure(font=self.FONT, default='disabled', **self.BORDER)

        tk.Label(w, text='Select the new VS properties', font=self.FONT).grid(row=0,column=1, sticky='we', columnspan=3,padx=5,pady=5)
        tk.Frame(w,height=2,bd=1,relief='sunken').grid(row=1, column=0, sticky='ew', columnspan=3, pady=3)
        # project 
        tk.Label(w, text='Project', font=self.FONTbold, width=12,anchor='e').grid(row=3,column=1,sticky='we')
        tk.Label(w, text='', font=self.FONT, width=10).grid(row=4,column=1,sticky='we',pady=5) # placeholder for entry

        self.prj_pull = OptionMenuFix(w,
               menubutton_width=30,
               menubutton_font=self.FONT,
               menu_font=self.FONT,
               menubutton_bd = 1, menubutton_highlightbackground = 'black',
               menubutton_borderwidth=1, menubutton_highlightcolor='black', 
               menubutton_highlightthickness = 1,
               menubutton_height=1,
               items = self.prj_list,
               initialitem=-1,
               command = self._setprjname)
        self.prj_pull.grid(row=3,column=2,sticky='we',padx=3)

        self.prj_new = Pmw.EntryField(w, value='', validate = {'validator' : hf.validateAscii, 'minstrict': 0}) #,
        self.prj_new.component('entry').configure(justify='left', font=self.FONT, bg='pink',width=33, **self.BORDER)

        # --------------------------------
        tk.Frame(w,height=2,bd=1,relief='sunken').grid(row=6, column=0, sticky='ew', columnspan=3, pady=3)
        
        # experiment
        tk.Label(w, text='Experiment', font=self.FONTbold, width=12,anchor='e').grid(row=7,column=1,sticky='we')
        tk.Label(w, text='', font=self.FONT, width=10).grid(row=8,column=1,sticky='we',pady=5) # placeholder for entry

        self.exp_pull = OptionMenuFix(w,labelpos='w',
                       menubutton_width=30,
                       menubutton_font=self.FONT,
                       menu_font=self.FONT,
               menubutton_bd = 1, menubutton_highlightbackground = 'black',
               menubutton_borderwidth=1, menubutton_highlightcolor='black', 
               menubutton_highlightthickness = 1,
               menubutton_height=1,
                       items=[self._new],
                       initialitem=-1,
                       command = self._setexpname)
        self.exp_pull.grid(row=7, column =2, sticky='we',padx=3)
        self.exp_new = Pmw.EntryField(w, value='', validate = {'validator' : hf.validateAscii, 'minstrict':0}) #,
        self.exp_new.component('entry').configure(justify='left', font=self.FONT, bg='pink',width=30, **self.BORDER)
        # initialize the interface with the projects
        self._setprjname()
        self.prj_pull.setvalue( self.prj_list[-1])

        # --------------------------------
        tk.Frame(w,height=2,bd=1,relief='sunken').grid(row=9, column=0, sticky='ew', columnspan=3, pady=3)
        # job tag 
        tk.Label(w, text='Optional jobs name tag', font=self.FONT).grid(row=10, column=1,columnspan=3,sticky='we',padx=5)
        self.tag_entry = Pmw.EntryField(w, value='', validate = hf.validateAsciiEmpty) #,
        self.tag_entry.component('entry').configure(justify='left', font=self.FONT, bg='white',width=30, **self.BORDER)
        self.tag_entry.grid(row=11,column=1, columnspan=3, sticky='we', padx=4,pady=4)

        self.win.bind('<Escape>', self.close)
        self.setSuggest()
        self.win.activate()


    def setSuggest(self):
        """ fill the submission with the suggestions"""
        if self.suggest == {}:
            return
        if 'prj' in self.suggest.keys():
            prj = self.suggest.pop('prj')
            self.prj_pull.setvalue(prj)
            self.prj_pull.invoke()
        if 'exp' in self.suggest.keys():
            exp = self.suggest.pop('exp')
            self.exp_pull.setvalue(exp)
            self.exp_pull.invoke()
        if 'tag' in self.suggest.keys():
            tag = self.suggest.pop('tag')
        else:
            tag = 'RESTARTED'
        self.tag_entry.setvalue(tag)




class ManageJobOverlaps(rb.RaccoonDefaultWidget):
    
    
    def __init__(self, parent, duplicates):
        rb.RaccoonDefaultWidget.__init__(self, parent)
        self.count = count
        #self.names = names
        self.remotenames = duplicates['server_duplicates']
        self.localnames = duplicates['local_duplicates']
        self.dialog = Pmw.Dialog(parent,
            buttons = ('Skip', 'Modify tag', 'Auto-rename', 'Overwrite', 'Cancel'),
            default_button = 'Modify tag',
            title = 'Jobs naming issues',
            command = self.execute)
        self.dialog.withdraw()
        d = self.dialog.interior()

        msg = ("The following jobs to be generated are going to "
               "overwrite jobs with the same names already "
               "present on the server.")
        pack_def = { 'side': 'top', 'anchor':'n', 'expand':0, 'fill':'x'}
        tk.Label(d, text = msg).pack(**pack_def)
        if len(self.remotenames):
            slb = Pmw.ScrolledListBox(d, listbox_font=self.FONT, items = self.remotenames,
                label_text = 'Server job names', labelpos='nw', label_font=self.FONTbold,
                listbox_selectmode='EXTENDED')
            slb.pack(side='left', anchor='w', expand='1', fill='both')
        if len(self.localnames):
            llb = Pmw.ScrolledListBox(d, listbox_font=self.FONT, items = self.localnames,
                label_text = 'Local job names', labelpos='nw', label_font=self.FONTbold,
                listbox_selectmode='EXTENDED')
            llb.pack(side='left', anchor='w', expand='1', fill='both')
        self.dialog.activate(globalMode=1)

        
    def execute(self, result):
        if result == 'Overwrite':
            t = 'Overwrite jobs'
            i = 'warning'
            m = ('Are you sure you want to overwrite %d jobs?' % len(names) )
            if not tmb.askyesno(parent=self.parent, title=t, icon=i, message=m):
                return
            choice == 'overwrite'
        elif result == 'Skip':
            choice = 'skip'
        elif result == 'Modify tag':
            choice = 'tag'
        elif result == 'Auto-rename':
            choice = 'rename'
        elif result == 'Cancel':
            choice = 'cancel'
        self.dialog.deactivate(choice)
            






   
from Pmv.moleculeViewer import MoleculeViewer
from DejaVu import Viewer
from ViewerFramework import VFGUI
#from DejaVu.Box import Box # THIS GOES INTO THE CODE THAT CREATES THE BOX
from MolKit import Read
from Tkinter import Toplevel

class EmbeddedMolViewer:

    def __init__(self, target=None,        # Tk container
                       name = '',
                       debug = 0,
                       customrc = None,
                ):
        self.debug = debug
        if self.debug:
            print "EmbeddedMolViewer __init__> ", target
        self.target = target
        self.PMV_win = Toplevel()
        self.PMV_win.withdraw()
        VFGUI.hasDnD2 = False
        print "\n\n--------[ Initializing PMV ]-----------"
        if customrc == None:
            customrc = '_rac_pmvrc'
            
        self.mv = MoleculeViewer(logMode = 'overwrite',  master=self.PMV_win, customizer=customrc,
                title='[embedded viewer]',
                withShell=False,
                verbose=False, gui = True, guiVisible=1)
        print "--------------[ done ]---------------"
        self.VIEWER = self.mv.GUI.VIEWER
        self.GUI = self.mv.GUI
        self.VIEWER.cameras[0].suspendRedraw = 1 # stop updating the main Pmv camera
        self._cam_counter = 0
        self.cameras = {} # 0 : { 'name' : 'grid box viewer' ,
                             #       'obj'  :  'pmv._xx_',
                             #       ''     :

        self.cameraslist = []    # 0, 1, 2 
                                 # keep the order in which cameras are added

        self._pmv_suspended_cams = []
        self.target = target

        if self.target:
            self.addCamera(target = self.targer, name=name)

    def _showPmv(self, event=None):
        if self.debug:
            print "_showPmv> called"
            print "ACTIVE CAMS", self.activeCams()
        self.VIEWER.cameras[0].suspendRedraw = 0 # enable updating the main Pmv camera
        self._pmv_suspended_cams = self.activeCams()

        self.deactivateCam()
        self.mv.GUI.ROOT.deiconify()
        self.VIEWER.cameras[0].Redraw()

    
    def _hidePmv(self, event=None):
        self.VIEWER.cameras[0].suspendRedraw = 1 # stop updating the main Pmv camera
        self.mv.GUI.ROOT.withdraw()
        self.activateCam(self._pmv_suspended_cams) # = self.activeCams()

    def addCamera(self,target,name = ''):
        if self.debug: print "adding another camera"

        c = self.VIEWER.AddCamera(master=target)       # the next ones will be added)
        cam = self.VIEWER.cameras[-1]
        self.cameraslist.append( cam )
        self.cameras[self._cam_counter] = { 'name':name, 'obj':c }
        self._cam_counter += 1
        return cam


    def delCamera(self, idx=None, name=None):
        if self.debug: print "delcamera", idx, name
        if idx == None and name == None:
            print "delete camera by name or by idx!"
            return
        if idx == 0 or len(self.VIEWER.cameras) == 0:
            print "cowardly refusing to delete the main Pmv camera...(returning)"
            return

        #print "CAMERAS _PRE"
        #print self.cameraslist
        #print self.cameras
        #print "============\n\n"
        self.VIEWER.DeleteCamera(idx)
        self.cameraslist.pop( self.cameraslist.index(idx) )
        del self.cameras[idx]
        #print "CAMERAS _PRE"
        #print self.cameraslist
        #print self.cameras
        #print "============\n\n"

 
    def activeCams(self):
        active = []
        for c in self.cameras.keys():
            if not self.cameras[c]['obj'].suspendRedraw:
                active.append(  self.cameras[c]['name'] )
        return active

    def cams(self):
        return [ self.cameras[i]['name'] for i in self.cameras.keys()] 
        
    def camByName(self,name):
        for c in self.cameras.keys():
            if self.cameras[c]['name'] == name:
                return self.cameras[c]['obj']


    def activateCam(self, namelist=[], only=1):
        # - hide/delete all res molecules
        # - reactivate this camera
        if self.debug: print "activatecam> namelist", namelist
        for c in self.cameras:
            if self.cameras[c]['name'] in namelist:
                self.cameras[c]['obj'].suspendRedraw = 0
                self.cameras[c]['obj'].Redraw()
            elif only:
                self.cameras[c]['obj'].suspendRedraw = 1

    def deactivateCam(self, namelist=[]):
        if len(namelist) == 0:
            namelist = self.cameras.keys()
        for c in namelist:
            self.cameras[c]['obj'].suspendRedraw = 1



if __name__ == '__main__':
    
    import Tkinter
    import Pmw
    root = Tkinter.Tk()
    print "\nINITIALIZING VIEWER...\n\n"
    print "PRINT ELLIPSIS", unicode(u"\u2026")
    vi = EmbeddedMolViewer(debug=1)
    print "[ done ]"
    root.bind('Escape', vi._showPmv)
    root.bind('Backspace', vi._hidePmv)
    group  = Pmw.Group(root, tag_text = 'Cameras')
    group.pack(expand=1,fill='both',side='bottom')
    Tkinter.Button(root, text='Add camera...', command = lambda : vi.addCamera(target=group.interior()) ).pack(anchor='w',side='left')
    Tkinter.Button(root, text='Del camera...', command = lambda : vi.delCamera(idx = vi.cameraslist[-1]) ).pack(anchor='w',side='left')
    root.mainloop()




#       
#           AutoDock | Raccoon2
#
#       Copyright 2013, Stefano Forli
#          Molecular Graphics Lab
#  
#     The Scripps Research Institute 
#           _  
#          (,)  T  h e
#         _/
#        (.)    S  c r i p p s
#          \_
#          (,)  R  e s e a r c h
#         ./  
#        ( )    I  n s t i t u t e
#         '
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import CADD.Raccoon2

# Raccoon stuff
import CADD.Raccoon2.HelperFunctionsN3P as hf
import RaccoonBasics as rb
import RaccoonEvents
import RaccoonLibraries
#import RaccoonServers

import os, Pmw, glob, time, fnmatch
from PmwOptionMenu import OptionMenu as OptionMenuFix
import Tkinter as tk
import TkTreectrl
import tkMessageBox as tmb
import tkFileDialog as tfd


from PIL import Image, ImageTk
import sys, time

# mgl modules
from mglutil.events import Event, EventHandler
from mglutil.util.callback import CallbackFunction #as cb



class LigandTab(rb.TabBase, rb.RaccoonDefaultWidget):
    
    def __init__(self, app, parent, debug=False):
        rb.TabBase.__init__(self, app, parent, debug = False)
        rb.RaccoonDefaultWidget.__init__(self, parent)
        self.resource = self.app.resource
        self.initIcons()
        self.app.eventManager.registerListener(RaccoonEvents.SetResourceEvent, self.handleResource)
        self.app.eventManager.registerListener(RaccoonEvents.ServerConnection, self.handleServerConnection)
        self.app.eventManager.registerListener(RaccoonEvents.ServerDisconnection, self.handleServerDisconnection)

    def initIcons(self):
        """ initialize the icons for the interface"""
        icon_path = CADD.Raccoon2.ICONPATH
        f = icon_path + os.sep + 'library_small.png'
        self._ICON_lib_small = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'library_small_rem.png'
        self._ICON_lib_small_rem = ImageTk.PhotoImage(Image.open(f))
        
        f = icon_path + os.sep + 'package.png'
        self._ICON_package = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'add.png'
        self._ICON_add = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'remove.png'
        self._ICON_remove = ImageTk.PhotoImage(Image.open(f))


    def handleResource(self, event=None):
        self.setResource(event.resource)


    def handleServerDisconnection(self, event=None):
        """ check for transfers?"""
        #print "GOT TO SERVER DISCONNECTION -->",
        self.handleServerConnection()

    def handleServerConnection(self, event=None):
        """ """
        #print "GOT TO SERVER CONNECTION (LIGAND)"
        self._populateliglibrarytoolbar()

    def setResource(self, resource):
        '''adapt the job manager panel to reflect currently selected resource'''
        if resource == self.resource:
            return
        self.resource = resource
        if resource == 'local':
            self.setLocalResource()
        elif resource == 'cluster':
            self.setClusterResource()
        elif resource == 'opal':
            self.setOpalResource()


    def setLocalResource(self):
        self._makelocliglibrary()
        pass

    def setClusterResource(self):
        #self.frame = tk.Frame(self.parent)
        self._makeclustliglibrary()

        self.frame.pack(expand=1, fill='both', anchor='n', side='top')
        #print "Raccoon GUI lig manager is now on :", self.app.resource

    def setOpalResource(self):
        self.resetFrame()
        pass


    def _makeclustliglibrary(self):
        """ """
        self.resetFrame()
        f = self.frame

        ### LOCAL library panel XXX
        #group = Pmw.Group(f, tag_text = 'Local libraries', tag_font = self.FONTbold)
        #tk.Label(group.interior(), text='NOTHING YET', bg='white').pack(expand=0, fill='x',padx=3)
        #group.pack(expand=0, fill='x',anchor='w', side='top',padx=5, pady=3)

        group = Pmw.Group(f, tag_text = 'Remote libraries', tag_font = self.FONTbold)
        group.pack(expand=1, fill='both',anchor='w', side='top',padx=5, pady=3)
        g = group.interior()

        tbar = tk.Frame(g)
        b1 = tk.Button(tbar, text = 'New', image=self._ICON_add, command = self._call_lib_manager,
            compound='left', font=self.FONTbold, **self.BORDER)
        b1.pack(anchor='w', side='left',padx=1)

        b2 = tk.Button(tbar, text = 'Remove', image=self._ICON_remove, command = self._call_lib_manager_del,
            compound='left', font=self.FONTbold, **self.BORDER)
        b2.pack(anchor='w', side='left',padx=0)

        # check for services upload/download
        # then disable the buttons
        tbar.pack(side='top', anchor='w',padx=2, pady=0)


        #g.pack(expand=1,fill='both',anchor='n', side='top')
       
        f = tk.Frame(g)

        tk.Label(f, text="Selected library :", font=self.FONT).pack(expand=0,anchor='w',side='left', padx=3,pady=1)
        self._selectedLib = tk.Label(f, text = '(none)', bg='white', anchor='w', font=self.FONTbold, **self.BORDER)
        self._selectedLib.pack(anchor='w',side = 'top', expand=1, fill='x', padx=3,pady=1)

        f.pack(expand=0,fill='both',anchor='n', side='top', padx=5, pady =3)



        self.ligLibPanel = TkTreectrl.ScrolledMultiListbox(g, bd=2)
        self.ligLibPanel.listbox.config(bg='white', fg='black', font=self.FONT,
                   columns = ('name', 'items', 'updated', 'format', 
                              'heavy min.', 'heavy max.', 'hba min.', 'hba max.',
                              'hbd min.', 'hbd max', 'mw min', 'mw max', 
                                'tors min', 'tors max', 'atypes') )
        self.ligLibPanel.pack(expand=1, fill='both')
        #self.ligLibPanel.listbox.bind('<ButtonRelease-1>', self._lib_callback)
        self.ligLibPanel.listbox.bind('<Button-3>', self.libselect) #_cb_ssh)
        
        if not self.app.server == None:
            self.dprint("UPDATE THE LIST OF LIBRARIES")


    def libselect(self, event=None):
        """ receive the event of a library selection and
            re-direct it to the resource-specific callback
        """
        self.ligLibPanel.listbox.event_generate('<Button-1>', x=event.x, y=event.y)
        if self.app.resource == 'local':
            self.libselect_cb_local() #event=event)
        elif self.app.resource =='cluster':
            self.libselect_cb_ssh() #event=event)
        elif self.app.resource == 'opal':
            self.libselect_cb_opal() #event=event)
        
    def libselect_cb_local(self): #, event=None):
        """ manage the selection of a library 
            for local submission
        """
        pass
        
    def libselect_cb_ssh(self, libname = None, filters = {}): #, event=None):
        """ manage the selection of a library for cluster submission
            TODO filters now are empty, so the entire library 
                is selected by default
        """
        if libname == None:
            try:
                s = self.ligLibPanel.listbox.curselection()[0]
                sel = self.ligLibPanel.listbox.get(s)[0]
            except:
                return
            libname = sel[0]
        #print "SELECTED LIB=[%s]" % libname
        self.dprint("SELECTED LIB=[%s]" % libname)
        # need ask the server for the lib
        lib = self.app.server.getLibrary(libname)
        if lib == None:
            t = 'Library selection'
            i = 'error'
            m = ('There was an error selecting the library.\n\n'
                 'The library "%s" does not exist on the server' % libname)
            tmb.showinfo(parent=self.ligLibPanel, title=t,message=m, icon=i)
            return False
        self.app.ligand_source = [ {'lib': lib, 'filters' : filters } ]
        self._selectedLib.config(text='%s' % libname)
        e = RaccoonEvents.UserInputRequirementUpdate('lig')
        self.app.eventManager.dispatchEvent(e)
        return True

        
    def libselect_cb_opal(self): #, event=None):
        """ manage the selection of a library for local submission
        """
        pass


    def libdeselect_cb(self):
        """ deselect the library"""
        self.app.ligand_source = []
        self._selectedLib.config(text='%s' % '(none)')
        e = RaccoonEvents.UserInputRequirementUpdate('lig')
        self.app.eventManager.dispatchEvent(e)


    def getSelectedLib(self):
        """ return the library selected for the docking"""
        sel = self.ligLibPanel.getselection

        return 

    def _call_lib_manager(self, event=None):
        """ """
        # test if there's a server
        if self.app.server == None:
            title ='No server selected'
            msg = ('Connect to a server in the Setup tab '
                   'to upload new libraries...')
            tmb.showinfo(parent = self.frame, title = title, message = msg)
            return
        # test if it is racconized
        if not self.app.server.properties['ready']:
            title ='Server is not ready'
            msg = ('The server has not been prepared ("racconized") yet.\n\n'
                   'Prepare it in the Setup tab and try again.')
            tmb.showinfo(parent = self.frame, title = title, message = msg)
            return
        # test if there is an upload service
        # XXX TODO XXX
        LibraryManagerWin(self.app, self.frame)

    def _call_lib_manager_del(self,event=None):
        # test if there's a server
        if self.app.server == None:
            title ='No server selected'
            msg = ('Connect to a server in the Setup tab '
                   'to remove libraries...')
            tmb.showinfo(parent = self.frame, title = title, message = msg)
            return
        # test if it is racconized
        if not self.app.server.properties['ready']:
            title ='Server is not ready'
            msg = ('The server has not been prepared ("racconized") yet.\n\n'
                   'Prepare it in the Setup tab and try again.')
            tmb.showinfo(parent = self.frame, title = title, message = msg)
            return
        if len(self.app.server.getLibraries()) == 0:
            title ='No libraries'
            msg = ('The server does not have libraries installed.')
            tmb.showinfo(parent = self.frame, title = title, message = msg)
            return
        try:
            s = self.ligLibPanel.listbox.curselection()[0]
            sel = self.ligLibPanel.listbox.get(s)[0]
        except:
            #print "Nothing selected, leaving..."
            return
        libname = sel[0]
        t = 'Delete library'
        m = ('The library "%s" is going to be deleted.\n\n'
               'All files on the remote server are going to be erased.\n\n'
               'Are you sure?') % libname
        w = 'warning'
        i = 'info'
        e = 'error'
        if not tmb.askyesno(parent=self.parent, title=t, message=m, icon =w):
            return
        # deselect library to be deleted if it is selected
        if len(self.app.ligand_source):
            if libname == self.app.ligand_source[0]['lib'].name:
                self.libdeselect_cb()
        reply = self.app.server.delLibrary(libname, delfiles=True)
        if reply:
            m = 'Operation completed'
            tmb.showinfo(parent = self.parent, title=t, message = m, icon = i)
        else:
            m = 'Operation was *NOT successful'
            tmb.showinfo(parent = self.parent, title=t, message = m, icon = e)
        self.app.setBusy()
        e = RaccoonEvents.ServerConnection()
        self.app.eventManager.dispatchEvent(e)
        self.app.setReady()

    def _makelocliglibrary(self):
        self.resetFrame()
        pass

    def _makeopalliglibrary(self):
        self.resetFrame()
        # option menu
        #self.libraryChooser = OptionMenuFix(f, labelpos='w', menubutton_width=40,
        #                    label_text = 'Library :',
        #                    menubutton_font = self.FONT,
        #                    menu_font = self.FONT,
        #                    command = self.chooseLibrary,
        #                    )
        #self.libraryChooser_NULL = '<no libraries>'
        #self.libraryChooser.pack()      
        pass

    def chooseLibrary(self):
        pass
        
    def _populateliglibrarytoolbar(self, event=None):
        """ add ligand libraries to the pulldown menu"""
        # delete ligand library container
        try:
            self.ligLibPanel.listbox.delete(0, 'end')
        except:
            print "BB> ERROR HERE", sys.exc_info()[1]
            pass
        # deselect library, if any selected
        self.libdeselect_cb()

        if self.app.server == None:
            return
        problematic = []
        for lib in self.app.server.getLibraries():
            try:
                d = lib.info
                p = lib.info['properties']
                date = "%d/%02d/%02d %02d:%02d:%02d" % tuple(d['date'])
                atypes = ",".join(sorted(p['atypes']))
                self.ligLibPanel.listbox.insert('END', lib.name(), d['count'], date, d['format'], 
                        p['heavy'][0],  p['heavy'][1],  p['hba'][0], p['hba'][1],  
                        p['hbd'][0],  p['hbd'][1], p['mw'][0],  p['mw'][1], 
                        p['tors'][0],  p['tors'][1], atypes )
            except:
                problematic.append(lib)
        if len(problematic):
            t = 'Library error'
            i = 'warning'
            m = ('%d libraries on the server had problems '
                 'and cannot be imported') % len(problematic)
            try:
                libnames = ''
                for l in problematic:
                    libnames += "- %s \n"  % l.name()
                m += ':\n\n' + libnames
            except:
                m += "."
            m += ('\n\nCheck config and data files in '
                  'the Raccoon directory on the server '
                  '(i.e. ~/raccooon/libraries/)')
            tmb.showinfo(parent=self.ligLibPanel.listbox, title=t, message=m, icon=i)


class LibraryManagerWin(rb.RaccoonDefaultWidget):
    """ class to create new libraries for receptors or ligands"""
    def __init__(self, app, parent=None, _type ='ligand', mode ='add' ):
        rb.RaccoonDefaultWidget.__init__(self, parent)
        self.app = app
        self._ligand_list = []
        self._register = {} # given a filename returns the name in LigBook
        self._type = _type  # ligand, receptor
        self.initIcons()
        self.makewin()

    def initIcons(self):
        """ initialize the icons for the interface"""
        icon_path = CADD.Raccoon2.ICONPATH
        f = icon_path + os.sep + 'package.png'
        self._ICON_package = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'system.png'
        self._ICON_sys = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'open.png'
        self._ICON_open = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'remove.png'
        self._ICON_remove = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'removex.png'
        self._ICON_remove = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'upload.png'
        self._ICON_upload = ImageTk.PhotoImage(Image.open(f))

        f = icon_path + os.sep + 'default.gif'
        self._ICON_default = ImageTk.PhotoImage(Image.open(f))


    def makewin(self):
        """ """
        self.win = Pmw.MegaToplevel(master = self.parent, title = 'Library Manager')
        topbar = tk.Frame(self.win.interior())
        tk.Label(topbar, text='Add files to create a new %s library...' % self._type, font=self.FONT, 
            compound='left', image = self._ICON_package).pack(anchor='w',side='left',pady=5, expand=1,fill='x')
        topbar.pack(anchor='n', side='top',expand=0, fill='x',padx=5, pady=1)
###        # settings button # XXX DISABLED
###        b = tk.Button(left_tbar, image= self._ICON_sys, **self.BORDER)
###        b.pack(anchor='n', side='top',expand=0, fill='x')
        self.center = Pmw.Group(self.win.interior(), tag_text = 'Local files [ 0 ]', tag_font=self.FONTbold)

        left_tbar = tk.Frame(self.center.interior())
        b = tk.Button(left_tbar, text='', image=self._ICON_open, **self.BORDER)
        b.pack(anchor='n', side='top',expand=0, fill='x')
        addlig_items = [ ['Add ligands'],
                         ['Import files...', 'normal', self.openfiles],
                         [],
                         ['Scan directory...', 'normal', self.open_dirs],
                         ['Scan directory (recursively)...', 'normal', self.open_dirs_recursive],
                        ]
        menu = rb.RacMenu(b, addlig_items, toolbar=left_tbar)

        b = tk.Button(left_tbar, text='[X]', image=self._ICON_remove, **self.BORDER)
        b.pack(anchor='n', side='top',expand=0, fill='x', pady=1)

        addlig_items = [ ['Remove ligands'],
                         ['Remove selected', 'normal', self.delete],
                         ['Remove all', 'normal', self.delete_all],
                        ]
        menu = rb.RacMenu(b, addlig_items, toolbar=left_tbar)

        left_tbar.pack(anchor='w', side='left', expand=0,fill='y',pady=2,padx=2)

        self.fmanager = Pmw.ScrolledListBox(self.center.interior(), listbox_font=self.FONT,
            listbox_selectmode='extended')
        self.fmanager.pack(expand=1, fill='both', anchor='n', side='top',padx=2, pady=1)
        delkey = CallbackFunction(self.delete, **{'nuke':False})
        self.fmanager.component('listbox').bind('<Delete>', delkey)

        self.uploadButton = tk.Button(self.center.interior(), text='Upload...', width=400,
            image=self._ICON_upload, compound='left', command=self.ask_upload,
            font=self.FONT, state='disabled', **self.BORDER)
        self.uploadButton.pack(anchor='s', side='bottom', expand=0, fill='x',padx=2, pady=2)

        self.center.pack(expand=1, fill='both', side='left',anchor='n',padx=3)

        right = tk.Frame(self.win.interior())
        b6 = tk.Button(right, text='Close', command=self.destroy, font=self.FONT, **self.BORDER)
        b6.pack(anchor='s', side='bottom',expand=1,fill='x')
        right.pack(anchor='e', side='bottom', expand=0, fill='y', pady=6,padx=5)
        self.win.activate()


    def _file_checker(self, flist):
        """query the Raccoon engine to check if files are acceptable
            and generate a minireport...?

            then proceed to adding them to the list
        """
        table = { 'rec' : 'receptors', 'lig' : 'ligands', 'flex' : 'flex.residues',
            'error' : 'errors/duplicates', 'result' : 'result files', 'empty' : 'empty file'}
        # response = self.app.engine.addLigandList(flist)
        if not flist:
            return 
        func = self.app.engine.addLigandList
        func_kwargs = {'filelist': flist }
        m = ('Checking %s potential ligands...' % len(flist))
        self.app.setBusy()
        progressWin = rb.ProgressDialogWindowTk(parent = self.win.interior(), 
                function = func, func_kwargs = func_kwargs, 
                title ='Importing ligands', message = m,
                operation = 'checking ligand files',
                image = self._ICON_open, autoclose=True, progresstype='percent')
        self.app.setReady()
        progressWin.start()
        response = progressWin.getOutput()
        if response == None:
            # response = {'accepted': [], 'rejected': [], 'receptor' : []}
            t = 'No ligands'
            m = ('No ligands have been found')
            i = 'info'
            tmb.showinfo(parent=self.parent, title=t, message=m, icon=i)
            return 
        if progressWin._STOP or (not progressWin._COMPLETED):
            if not len(response['accepted']):
                return
            t = 'Importing ligands'
            m = ('The ligand checking has not been completed.\n'
                 'Do you want to process %d ligands found so far?')
            m = m % len(response['accepted'])
            i = 'info'
            if not tmb.askyesno(parent=self.parent, title=t, message=m, icon=i):
                return 
        accepted = response['accepted']
        self.addfiles(accepted)
        for f in accepted:
            self._register[f[0]] = f[1] # RaccoonEngine.addLigandList -> accepted.append(f, response['name'])
        rejected = response['rejected']
        def close(event=None):
            win.destroy()
        win = Pmw.Dialog(self.win.interior(), title='Report', 
            buttons = ('Close',), command = close)
        bbox = win.component('buttonbox')
        for i in range(bbox.numbuttons()):
            bbox.button(i).configure(font=self.FONT, default='disabled', **self.BORDER)
        w = tk.Frame(win.interior(), bg='white', **self.BORDER)
        tk.Label(w, text='Accepted :', justify='left',anchor='e', font=self.FONTbold,
            bg='white').grid(row = 1, column = 1, sticky='we')
        tk.Label(w, text=len(accepted), justify='left', font=self.FONT, bg='white', 
            anchor='e').grid(row = 1, column = 2, sticky='we')
        tk.Frame(w,height=2,width=1,bd=1,relief='sunken').grid(row=3,column=0,sticky='we', 
            columnspan=13)
        tk.Label(w, text='Problematic ', justify='left',anchor='e', font=self.FONTbold, 
            bg='white').grid(row = 4, column = 1, sticky='we')
        r = 5
        for e in rejected.keys():
            msg = table.get(e, 'unknown')
            tk.Label(w, text=msg+" :", justify='left', anchor='e',font=self.FONT, 
                bg='white').grid(row = r, column = 1, sticky='we')
            tk.Label(w, text=len(rejected[e]), justify='left', font=self.FONT, 
                bg='white').grid(row = r, column = 2, sticky='we')
            r+=1
        w.pack(expand=1,fill='both', padx=10, pady=10, ipadx=5, ipady=5)
        win.activate() #$geometry='centerscreenalways')

    def addfiles(self, flist):
        """ add files to the lisbox and the self._ligand_list"""
        fnames = [ x[0] for x in flist ]
        for f in fnames:
            self._ligand_list.append(f)
            #self.fmanager.insert('end', f)
            self._update_count()
        self.fmanager.setlist(self._ligand_list)

    def openfiles(self, event=None):
        """ add files"""
        title = 'Select one or more %s files...' % self._type

        ## filetypes = [("Supported ligand formats", ("*.pdbqt", "*.mol2", "*.pdb" )), # XXX include case-sensitive alternate
        ##             ("PDBQT", "*.pdbqt"), ("PDB", "*.pdb"), ("Mol2", "*.mol2"), 
        ##             ("Any file type...", "*")]        
        
        filetypes = [("Supported ligand formats", ("*.pdbqt","*.PDBQT" )),
                     ("PDBQT", ("*.pdbqt", "*.PDBQT")), ("Any file type...", "*")]        
        files = tfd.askopenfilename(parent=self.win.interior(), title = title,
            filetypes = filetypes, multiple = 1)
        files = self._listfixer(files)
        if len(files):
            self._file_checker(files)

    def open_dirs_recursive(self, event=None):
        """call self.add_dirs with recursive mode"""
        self.open_dirs(recursive=True)

    def open_dirs(self,event=None, recursive=False):  
        """ scan dir(s) optionally recursively"""
        title = 'Select a directory to scan'
        if recursive:
            title += ' recursively'
        dirname = tfd.askdirectory(parent=self.win.interior(), title = title, mustexist=True)
        if dirname:
            # b = tk.Button(left_tbar, text='', image=self._ICON_open, **self.BORDER)
            self.app.setBusy()
            m = ('Scanning path: %s' % dirname)
            files = []
            func = hf.pathToList
            func_kwargs = {'path':dirname, 'pattern' :'*.pdbqt', 'recursive' : recursive}
            progressWin = rb.ProgressDialogWindowTk(parent= self.win.interior(),
                function = func, func_kwargs = func_kwargs, title ='Searching for ligands', 
                    message = m, operation = 'ligands scanning',image = self._ICON_open, autoclose=True)
            progressWin.start()
            self.app.setReady()
            files = progressWin.getResults()
            if files == []:
                t = 'Searching for ligands'
                m = ('No files found.')
                i = 'info'
                tmb.showinfo(parent=self.win.interior(), title=t, message=m, icon=i)
                return 
            if progressWin._STOP or not progressWin._COMPLETED:
                t = 'Searching for ligands'
                m = ('The ligand search has not been completed.\n'
                     'Do you want to process %d files found so far?')
                m = m % len(files)
                i = 'info'
                if not tmb.askyesno(parent=self.win.interior(), title=t, message=m, icon=i):
                    return 
            self._file_checker(files)
    
    def delete(self, event=None, nuke=False):
        """ delete selected or all ligands"""
        if nuke:
            self.fmanager.setlist(())
            del self._ligand_list[:]
            self.app.engine.removeLigands()
            del self._register
            self._register = {}
        else:
            try:
                lig_sel = self.fmanager.getcurselection()# [0]
            except:
                return
            for lig in lig_sel:
                idx = self._ligand_list.index(lig)
                self._ligand_list.pop(idx)
                name = self._register.pop(lig)
                self.app.engine.removeLigands([name])
            self.fmanager.setlist(self._ligand_list)
        self._update_count()
    
    def delete_all(self, event=None):
        if len(self._ligand_list) == 0:
            return
        msg = ('%d ligands are going to be removed' 
               'from the list.\n\nContinue?') %  len(self._ligand_list)
        title = 'Warning'
        icon = 'warning'
        if not tmb.askyesno(parent = self.win.interior(), title=title, message = msg, icon=icon):
            return
        self.delete(nuke=True)


    def _update_count(self, event=None):
        """ set the label of number of items"""
        self.center.configure(tag_text = 'Local files [%d]' % len(self._ligand_list))
        if len(self._ligand_list):
            self.uploadButton.configure(state='normal')
        else:
            self.uploadButton.configure(state='disabled')

    def destroy(self, event=None):
        """
        """
        if len(self._ligand_list) > 0:
            msg = ('There are %d ligands selected to be uploaded.\n\n'
                    'If the Ligand manager will be closed, they '
                    'will not be uploaded to the server.\n\nContinue?') % len(self._ligand_list)
            if not tmb.askyesno(parent=self.win.interior(), title='Warning', message = msg, icon ='warning'):
                return
        self.app.engine.removeLigands()
        self.win.deactivate()


    def ask_upload(self):
        """upload to remote location"""
        if len(self._ligand_list) == 0:
            return

        def close(event=None):
            if event == 'Start':
                server = self.app.server
                # check if lib name is in use
                libname = self._lname.getvalue()
                lib = server.getLibrary(libname)

                if lib: # and lib.info['type'] == self.info['type']:
                    t = 'Name already in use'
                    m = ('There is already a %s library with the same name.\n\n'
                           'Choose a different name and try again.') % self._type
                    i = 'warning'
                    tmb.showinfo(parent=win.interior(), title=t, message=m, icon=i)
                    return
                # check directory name
                dirname = self._dname.getvalue()
                remotepath = self._remotepath.getvalue()
                fullpath = remotepath + '/' + dirname
                if server.ssh.exists(fullpath):
                    t = 'Directory exist'
                    m = ('There is already a directory with the same in the'
                         'remote path. Select a different dirname '
                           'and try again.')
                    i = 'warning'
                    tmb.showinfo(parent=win.interior(), title=t, message=m, icon=i)
                    self.app.setReady()
                    return
                # comments, if any
                comment = self._comment.get('1.0', 'end').strip()
                win.destroy()
                self.app.setBusy()
                self.run_upload(name = libname, 
                    dirname = dirname, 
                    remotepath = remotepath, 
                    comment = comment)
                # once files habe been uploaded, clean up the engine from used files
                self.app.engine.removeLigands()
                self.app.setReady()
            else:
                win.destroy()
                self.app.setReady()

        def makedirname(event=None):
            """ ButtonRelease """
            chars = hf._allowedfilenamechars()
            text = self._lname.getvalue()
            if not len(text):
                self._dname.setvalue('')

            valid = [ x for x in text if x in hf._allowedfilenamechars()]
            valid = "".join(valid).lower() 
            self._dname.setvalue(valid)

        def setdefaultremote(event=None):
            self._remotepath.setvalue( self.app.server.getLibraryPath() )

        #settings = { 'split' : 1000, 
        #    }
        win = Pmw.Dialog(self.win.interior(), title = 'Upload library', buttons = ('Start','Cancel'), 
                command=close)
        w = win.interior()
        bbox = win.component('buttonbox')
        for i in range(bbox.numbuttons()):
            bbox.button(i).configure(font=self.FONT, default='disabled', **self.BORDER)
        w.pack_configure(ipadx=10, ipady=10)

        tk.Label(w, text=' ', image=self._ICON_upload, anchor='n', font=self.FONTbold).grid(row=0,
            column=0, sticky='we', columnspan = 3)
        tk.Label(w, text='Library name', anchor='e', font=self.FONT).grid(row=1, column=1, sticky='e')
        self._lname = Pmw.EntryField( w, validate = {'min':1,'minstrict':0}, entry_width=30, 
            entry_font=self.FONT)
        self._lname.component('entry').bind('<KeyRelease>', makedirname)
        self._lname.grid(row=1, column=2,sticky='w',columnspan=1)

        tk.Label(w, text='Remote path', anchor='e', font=self.FONT).grid(row=3, column=1, sticky='e')
        self._remotepath = Pmw.EntryField( w, validate = {'min':1,'minstrict':0}, 
            entry_width=30, entry_font=self.FONT)
        self._remotepath.grid(row=3, column=2,sticky='w',columnspan=1)
        tk.Button(w, text='D', command=setdefaultremote,image=self._ICON_default,   
            **self.BORDER).grid(row=3, column=3, sticky='w')

        # FIXME probably useless, this should be hidden to the user!
        #tk.Label(w, text='Directory name', anchor='e').grid(row=5, column=1, sticky='e', 
        #    entry_font=self.FONT)
        self._dname = Pmw.EntryField( w, validate = hf.validateFname, entry_width=30)
        #self._dname.grid(row=5,column=2, sticky='w', columnspan=1)
        tk.Label(w, text='Comments', anchor='e', font=self.FONT).grid(row=6, column=1, sticky='e')
        self._comment = hf.TextCopyPaste( w, bg='white', height=10, width=30, font=self.FONT)
        self._comment.grid(row=6,column=2, sticky='w')
        setdefaultremote()
        win.activate() # geometry='centerscreenalways' )

    def run_upload(self, name, dirname, remotepath, comment=None):
        """ """
        #    self.run_upload(name = libname, dirname = dirname, remotepath = remotepath)
        def _go(event=None):
            #server.debug = True
            #libobj.debug= True
            self.app.setBusy()
            if comment:
                libobj.info['comment'] = comment
            libobj.upload( dirname, remotepath, autoregister=0, bg = True)
            self.RUNNING = True
            self._stopclosebutton.configure(text= 'Stop', command=stop)
            status = server.transfer.progress(percent=1)

            while status < 100.:
                if server.transfer._STOP: break
                status = server.transfer.progress(percent=1)
                pc_var.set(status)
                self.bar.update()
                time.sleep(0.2)
            self.RUNNING = False

            libobj.debug = True # DEBUG 
            if server.transfer._status['completed']:
                # EVENT trigger the event refresh of server list
                #idxfile = remotepath + '/'+ dirname +'/'+'library.db'
                #idxfile = 'library.db'
                self._statuslab1.configure(text='Generating index file')
                self._statuslab2.configure(text='...saving...')
                libobj.saveIndex()
                self._statuslab1.configure(text='Registering library to Library Index')
                self._statuslab2.configure(text='...ongoing...')
                libobj.register()
                self._statuslab1.configure(text='\nLibrary upload completed!')
                self._statuslab2.configure(text=' ')
                del self._ligand_list[:]
                self.delete(nuke=True)
                e = RaccoonEvents.ServerConnection()
                self.app.eventManager.dispatchEvent(e)
            else:
                error = server.transfer_status['error']
                self._statuslab1.configure(text='Error trasnferring files!')
                self._statuslab2.configure(text=error)
            self._stopclosebutton.configure(text= 'Close', command=forceclose)
            self.app.setReady()

        def close(event=None):
            if event == None:
                #print "probably clicking on the corner..."
                return
            win.destroy()

        def stop(event=None):
            t = 'Stop transfer'
            m = 'The transfer is incomplete: do you want to stop it?'
            if not tmb.askyesno(parent = win.interior(), title=t, message=m):
                return
            server.transfer.stop()
            self._stopclosebutton.configure(text= 'Close', command=close)

        forceclose = CallbackFunction(close, True)


        ## print ">> adding ligands to lib object"
        # threaded files scanning...
        server = self.app.server
        libobj = RaccoonLibraries.LigandLibrary( server, name = name )
        libobj.options['split'] = 1000
        func = libobj.addLigands
        func_kwargs = {'ligands' : self._ligand_list }
        m = ('Extracting ligand properties...' )
        self.app.setBusy()
        progressWin = rb.ProgressDialogWindowTk(parent = self.parent,
                function = func, func_kwargs = func_kwargs, 
                title ='Ligand library', message = m, 
                operation = 'ligand properties extraction',
                image = self._ICON_open, autoclose=True, 
                progresstype='percent')
        progressWin.start()
        problematic = progressWin.getResults()
        if problematic == None: problematic = [] # normalize data structure
        self.app.setReady()
        #if response == None:
        #libobj.addLigands(self._ligand_list) # XXX TODO FIXME 

        win = Pmw.MegaToplevel(master = self.win.interior(), title = 'Library manager')
        win.userdeletefunc(close)
        win.component('hull').configure(width=600, height=400)
        
        fullpath = remotepath + '/' + dirname
        self._statuslab1 = tk.Label(win.interior(), text = 'Transferring files...', font=self.FONTbold)
        self._statuslab1.pack(expand=0, fill='x', anchor='w', side='top')
        self._statuslab2 = tk.Label(win.interior(), text = 'Destination: %s' % fullpath, font=self.FONT)
        self._statuslab2.pack(expand=0, fill='x', anchor='w', side='top')
        pc_var = tk.DoubleVar(value=0.)
        self.bar = hf.ProgressBar(win.interior(), pc_var)
        self.bar.pack(expand=0, fill='none', anchor='n', side='top',padx=5, pady=3)
        tk.Frame(win.interior(),height=2,width=1,bd=1,relief='sunken').pack(expand=0, fill='x', anchor='w', side='top')
        self._stopclosebutton = tk.Button(win.interior(), text='STOP', font=self.FONT)
        self._stopclosebutton.pack(expand=0, fill='none', anchor='s', side='bottom')

        win.interior().after(50, _go)
        win.activate() # geometry='centerscreenalways' )
            

